import globals
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.constant import HealthCheckConstant, ConfigurationKeyConstant
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod
from python_framework.api.src.dto import ActuatorHealthDto


@Controller(url=HealthCheckConstant.URI, tag='HealthCheck', description='HealthCheck controller')
class ActuatorHealthController:

    @ControllerMethod(
        responseClass=ActuatorHealthDto.ActuatorHealthResponseDto
        # , logResponse=globals.getGlobalsInstance().getApiSetting(ConfigurationKeyConstant.HEALTH_CHECK_LOG_RESPONSE)
    )
    def get(self):
        return self.service.actuatorHealth.getStatus(), HttpStatus.OK



from flask import Flask
from flask_restful import Api
from flask_cors import CORS
from python_helper import Constant as c
from python_helper import log, Function, ReflectionHelper, ObjectHelper, StringHelper
import globals
from python_framework.api.src.constant import ControllerConstant
from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.constant import StaticConstant
from python_framework.api.src.constant import HealthCheckConstant
from python_framework.api.src.util import FlaskUtil
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.service import SqlAlchemyProxy
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service import ApiKeyManager
from python_framework.api.src.service import SecurityManager
from python_framework.api.src.service import SchedulerManager
from python_framework.api.src.service.openapi import OpenApiManager


DOT_PY = '.py'


def getPythonFrameworkResourceByType(resourceType):
    return FlaskManager.PYTHON_FRAMEWORK_RESOURCE_NAME_DICTIONARY.get(resourceType, [])


def isNotPythonFrameworkApiInstance(apiInstance):
    return apiInstance.globals.apiName not in FlaskManager.PYTHON_FRAMEWORK_INTERNAL_MODULE_NAME_LIST


def isFromPythonFramework(apiInstance, resourceType, resourceName):
    return not (resourceName in getPythonFrameworkResourceByType(resourceType) and isNotPythonFrameworkApiInstance(apiInstance))


def getResourceModuleNameAjusted(apiInstance, resourceType, resourceName):
    return resourceName if isFromPythonFramework(apiInstance, resourceType, resourceName) else FlaskManager.PYTHON_FRAMEWORK_MODULE_NAME


def getResourceNameAjusted(apiInstance, resourceType, resourceName):
    return resourceName if isFromPythonFramework(apiInstance, resourceType, resourceName) else f'{resourceName}{c.DOT}{resourceName}'


def isControllerResourceName(resourceName):
    return FlaskManager.KW_CONTROLLER_RESOURCE == resourceName[-len(FlaskManager.KW_CONTROLLER_RESOURCE):]


@Function
def getResourceName(resourceFileName):
    return resourceFileName.split(DOT_PY)[0]


@Function
def isResourceType(resourceFileName,resourceType):
    splitedResourceFileName = resourceFileName.split(resourceType)
    return len(splitedResourceFileName)>1 and splitedResourceFileName[1] == DOT_PY


@Function
def getResourceNameList(apiTree, resourceType):
    resourceNameList = []
    if apiTree or type(apiTree).__name__ == c.DICT:
        for package,subPackageTree in apiTree.items():
            if isResourceType(package, resourceType):
                resourceNameList.append(getResourceName(package))
            resourceNameList += getResourceNameList(
                subPackageTree,
                resourceType
            )
    return resourceNameList


@Function
def getControllerNameList(controllerName):
    # controllerNameList = [
    #     name for name in dir(globals.importModule(controllerName)) if not name.startswith(c.UNDERSCORE) and name.endswith(FlaskManager.KW_CONTROLLER_RESOURCE)
    # ]
    controllerNameList = [controllerName]
    for controllerType in ControllerConstant.CONTROLLER_TYPE_LIST:
        controllerNameList.append(f'{controllerName[:-len(FlaskManager.KW_CONTROLLER_RESOURCE)]}{controllerType}{FlaskManager.KW_CONTROLLER_RESOURCE}')
    return controllerNameList


@Function
def getControllerList(resourceName, resourceModuleName):
    controllerNameList = getControllerNameList(resourceName)
    importedControllerList = []
    for controllerName in controllerNameList:
        resource = globals.importResource(controllerName, resourceModuleName=resourceModuleName)
        if ObjectHelper.isNotNone(resource):
            importedControllerList.append(resource)
    if 0 == len(importedControllerList):
        raise Exception(f'Not possible to import {resourceName} controller')
    return importedControllerList


@Function
def getResourceList(apiInstance, resourceType):
    resourceNameList = getResourceNameList(
        apiInstance.globals.apiTree[apiInstance.globals.apiPackage],
        resourceType
    )
    if isNotPythonFrameworkApiInstance(apiInstance):
        resourceNameList += getPythonFrameworkResourceByType(resourceType)
    resourceList = []
    for resourceName in resourceNameList:
        resource = None
        ajustedResourceName = getResourceNameAjusted(apiInstance, resourceType, resourceName)
        ajustedResourceModuleName = getResourceModuleNameAjusted(apiInstance, resourceType, resourceName)
        if isControllerResourceName(resourceName):
            resource = getControllerList(ajustedResourceName, ajustedResourceModuleName)
        else:
            resource = globals.importResource(ajustedResourceName, resourceModuleName=ajustedResourceModuleName, required=True)
        if ObjectHelper.isEmpty(resource):
            raise Exception(f'Error while importing {ajustedResourceName} resource from {ajustedResourceModuleName} module. Resource not found.')
        elif ObjectHelper.isList(resource):
            resourceList += resource
        else:
            resourceList.append(resource)
    return resourceList


@Function
def fixInitializeKwargs(initializeKwargs):
    if ObjectHelper.isNotEmpty(initializeKwargs):
        for key in [
            'static_url_path',
            'static_folder',
            'template_folder'
        ]:
            if key in initializeKwargs:
                initializeKwargs.remove(key)


@Function
def getBaseUrl(globalsInstance):
    return globalsInstance.getSetting(ConfigurationKeyConstant.API_SERVER_BASE_URL)


@Function
def getStaticBaseUrl(staticPackage, givenBaseStaticUrl, globalsInstance):
    if ObjectHelper.isNotEmpty(givenBaseStaticUrl):
        return givenBaseStaticUrl
    return f'{getBaseUrl(globalsInstance)}{c.FOWARD_SLASH}{staticPackage}'


@Function
def addGlobalsTo(apiInstance, globalsInstance=None):
    FlaskUtil.validateFlaskApi(apiInstance)
    apiInstance.globals = globalsInstance if ObjectHelper.isNotNone(globalsInstance) else FlaskUtil.getGlobals()
    apiInstance.globals.api = apiInstance
    apiInstance.bindResource = FlaskManager.bindResource


@Function
def getApiStaticUrl(app=None, api=None):
    if ObjectHelper.isNone(api):
        api = app.api
    return f'{api.documentationUrl[:-(len(api.baseUrl)+len(OpenApiManager.DOCUMENTATION_ENDPOINT))]}{api.baseStaticUrl}'


@Function
def initialize(
    rootName,
    refferenceModel,
    managerList = None,
    staticPackage = StaticConstant.KW_STATIC_FOLDER,
    viewPackage = StaticConstant.KW_VIEW_FOLDER,
    staticUrl = None,
    **kwargs
):
    fixInitializeKwargs(kwargs)
    if ObjectHelper.isNone(managerList):
        managerList = []

    globalsInstance = FlaskUtil.getGlobals()

    baseStaticUrl = getStaticBaseUrl(staticPackage, staticUrl, globalsInstance)

    app = Flask(
        rootName,
        static_folder = staticPackage,
        template_folder = viewPackage,
        static_url_path = baseStaticUrl,
        **kwargs
    )
    api = Api(app)
    api.app = app
    api.app.api = api
    api.managerList = managerList

    addGlobalsTo(api, globalsInstance=globalsInstance)

    api.scheme = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_SCHEME)
    api.host = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_HOST)
    api.port = globalsInstance.getApiSetting(ConfigurationKeyConstant.API_SERVER_PORT)
    api.baseUrl = getBaseUrl(globalsInstance)
    api.baseStaticUrl = baseStaticUrl
    api.internalUrl = FlaskManager.getInternalUrl(api)

    api.cors = CORS(app, resources={f"{api.baseUrl}/{c.ASTERISK}": {'origins': c.ASTERISK}})
    api.cors.api = api

    OpenApiManager.newDocumentation(api, app)
    SqlAlchemyProxy.addResource(api, app, baseModel=refferenceModel, echo=False)
    SchedulerManager.addResource(api, app)
    SessionManager.addResource(api, app)
    ApiKeyManager.addResource(api, app)
    SecurityManager.addResource(api, app)
    for manager in api.managerList:
        manager.addResource(api, app)
    addFlaskApiResources(*[api, app, *[getResourceList(api, resourceType) for resourceType in FlaskManager.KW_RESOURCE_LIST]])
    for manager in api.managerList:
        manager.onHttpRequestCompletion(api, app)
    SessionManager.onHttpRequestCompletion(api, app)
    ApiKeyManager.onHttpRequestCompletion(api, app)
    SecurityManager.onHttpRequestCompletion(api, app)
    SchedulerManager.onHttpRequestCompletion(api, app)
    SqlAlchemyProxy.onHttpRequestCompletion(api, app)

    api.documentationUrl = OpenApiManager.getDocumentationUrl(api)
    api.healthCheckUrl = f'{api.documentationUrl[:-len(OpenApiManager.DOCUMENTATION_ENDPOINT)]}{HealthCheckConstant.URI}'
    api.staticUrl = getApiStaticUrl(api=api)

    return app


@Function
def addControllerListTo(apiInstance, controllerList):
    for controller in controllerList:
        OpenApiManager.addControllerDocumentation(controller, apiInstance)
        mainUrl = f'{apiInstance.baseUrl}{controller.url}'
        urlList = [mainUrl]
        infoList = [f'Controller: {mainUrl}']
        controllerMethodList = ReflectionHelper.getAttributePointerList(controller)
        for controllerMethod in controllerMethodList:
            if ReflectionHelper.hasAttributeOrMethod(controllerMethod, FlaskManager.KW_URL) and ObjectHelper.isNotEmpty(controllerMethod.url):
                controllerUrl = f'{mainUrl}{controllerMethod.url}'
                if controllerUrl not in urlList:
                    urlList.append(controllerUrl)
                    infoList.append(f'{c.TAB}{ReflectionHelper.getName(controllerMethod)}: {controllerUrl}')
                # subUrlList = controllerMethod.url.split(c.SLASH)
                # concatenatedSubUrl = c.NOTHING
                # for subUrl in subUrlList:
                #     if subUrl:
                #         concatenatedSubUrl += f'{c.SLASH}{subUrl}'
                #         if c.LESSER == subUrl[0] and c.BIGGER == subUrl[-1]:
                #             newUrl = f'{apiInstance.baseUrl}{controller.url}{concatenatedSubUrl}'
                #             if not newUrl in urlList:
                #                 urlList.append(newUrl)
                OpenApiManager.addEndPointDocumentation(controllerUrl, controllerMethod, controller, apiInstance)
        log.debug(addControllerListTo, f'{controller.url} -> {StringHelper.prettyPython(infoList)}')
        apiInstance.add_resource(controller, *urlList)


@Function
def addServiceListTo(apiInstance,serviceList):
    for service in serviceList:
        apiInstance.bindResource(apiInstance,service())


@Function
def addSchedulerListTo(apiInstance,schedulerList):
    for scheduler in schedulerList:
        apiInstance.bindResource(apiInstance,scheduler())


@Function
def addClientListTo(apiInstance,clientList):
    for client in clientList:
        apiInstance.bindResource(apiInstance,client())


@Function
def addListenerTo(apiInstance,listenerList):
    for listener in listenerList:
        apiInstance.bindResource(apiInstance,listener())


@Function
def addEmitterTo(apiInstance, emitterList):
    for emitter in emitterList:
        apiInstance.bindResource(apiInstance,emitter())


@Function
def addRepositoryTo(apiInstance, repositoryList):
    for repository in repositoryList:
        apiInstance.bindResource(apiInstance,repository())


@Function
def addValidatorListTo(apiInstance,validatorList):
    for validator in validatorList:
        apiInstance.bindResource(apiInstance,validator())


def addMapperListTo(apiInstance,mapperList):
    for mapper in mapperList:
        apiInstance.bindResource(apiInstance,mapper())


@Function
def addHelperListTo(apiInstance,helperList):
    for helper in helperList:
        apiInstance.bindResource(apiInstance,helper())


@Function
def addConverterListTo(apiInstance,converterList):
    for converter in converterList:
        apiInstance.bindResource(apiInstance,converter())


class FlaskResource:
    ...


@Function
def addResourceAttibutes(apiInstance):
    ReflectionHelper.setAttributeOrMethod(apiInstance, FlaskManager.KW_RESOURCE, FlaskResource())
    for resourceName in FlaskManager.KW_RESOURCE_LIST:
        ReflectionHelper.setAttributeOrMethod(apiInstance.resource, f'{resourceName[0].lower()}{resourceName[1:]}', FlaskResource())


@Function
def addFlaskApiResources(
        apiInstance,
        appInstance,
        controllerList,
        schedulerList,
        serviceList,
        clientList,
        listenerList,
        emitterList,
        repositoryList,
        validatorList,
        mapperList,
        helperList,
        converterList,
        *args
    ):
    addResourceAttibutes(apiInstance)
    addRepositoryTo(apiInstance, repositoryList)
    addSchedulerListTo(apiInstance, schedulerList)
    addClientListTo(apiInstance, clientList)
    addListenerTo(apiInstance, listenerList)
    addEmitterTo(apiInstance, emitterList)
    addServiceListTo(apiInstance, serviceList)
    addControllerListTo(apiInstance, controllerList)
    addValidatorListTo(apiInstance, validatorList)
    addMapperListTo(apiInstance, mapperList)
    addHelperListTo(apiInstance, helperList)
    addConverterListTo(apiInstance, converterList)
    SqlAlchemyProxy.initialize(apiInstance, appInstance)
    SchedulerManager.initialize(apiInstance, appInstance)
    SecurityManager.initialize(apiInstance, appInstance)
    ApiKeyManager.initialize(apiInstance, appInstance)
    SessionManager.initialize(apiInstance, appInstance)
    for manager in apiInstance.managerList:
        manager.initialize(apiInstance, appInstance)
    OpenApiManager.addSwagger(apiInstance, appInstance)



from python_helper import EnvironmentHelper
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.annotation.client.HttpClientAnnotation import HttpClient, HttpClientMethod
from python_framework.api.src.util import Serializer
from python_framework.api.src.dto import ActuatorHealthDto

import ClientTestDto


BASE_CONTROLLER_URL = f'/exception/test/{EnvironmentHelper.get("URL_VARIANT")}'


@HttpClient(url='http://localhost:5022/client-test-api')
class ExceptionTestClient:

    @HttpClientMethod(
        url = f'{BASE_CONTROLLER_URL}',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [str],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logResponse = True,
        logRequest = True
    )
    def getMethod(self, uri, headers=None, params=None):
        return self.get(
            additionalUrl = uri,
            params = Serializer.getObjectAsDictionary(params),
            headers = {'Some-Cool-Header': 'cool-value', **Serializer.getObjectAsDictionary(headers)}
        )

    @HttpClientMethod(
        url = f'{BASE_CONTROLLER_URL}',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [str, ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logResponse = True,
        logRequest = True
    )
    def postMethod(self, uri, dto, headers=None, params=None):
        return self.post(
            body = Serializer.getObjectAsDictionary(dto),
            additionalUrl = uri,
            params = Serializer.getObjectAsDictionary(params),
            headers = {'Some-Cool-Header': 'cool-value', **Serializer.getObjectAsDictionary(headers)}
        )

    @HttpClientMethod(
        url = f'{BASE_CONTROLLER_URL}',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [str, ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logResponse = True,
        logRequest = True
    )
    def putMethod(self, uri, dto, headers=None, params=None):
        return self.put(
            body = Serializer.getObjectAsDictionary(dto),
            additionalUrl = uri,
            params = Serializer.getObjectAsDictionary(params),
            headers = {'Some-Cool-Header': 'cool-value', **Serializer.getObjectAsDictionary(headers)}
        )

    @HttpClientMethod(
        url = f'{BASE_CONTROLLER_URL}',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [str, ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logResponse = True,
        logRequest = True
    )
    def patchMethod(self, uri, dto, headers=None, params=None):
        return self.patch(
            body = Serializer.getObjectAsDictionary(dto),
            additionalUrl = uri,
            params = Serializer.getObjectAsDictionary(params),
            headers = {'Some-Cool-Header': 'cool-value', **Serializer.getObjectAsDictionary(headers)}
        )

    @HttpClientMethod(
        url = f'{BASE_CONTROLLER_URL}',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [str, ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logResponse = True,
        logRequest = True
    )
    def deleteMethod(self, uri, dto, headers=None, params=None):
        return self.delete(
            body = Serializer.getObjectAsDictionary(dto),
            additionalUrl = uri,
            params = Serializer.getObjectAsDictionary(params),
            headers = {'Some-Cool-Header': 'cool-value', **Serializer.getObjectAsDictionary(headers)}
        )



from python_helper import Test
from python_framework.api.src.service.flask import FlaskManager

@Test()
def KW_RESOURCE_LIST_isInCorrectOrder() :
    # arrange
    resourceNameCorrectOrder = [
        FlaskManager.KW_CONTROLLER_RESOURCE,
        FlaskManager.KW_SCHEDULER_RESOURCE,
        FlaskManager.KW_SERVICE_RESOURCE,
        FlaskManager.KW_CLIENT_RESOURCE,
        FlaskManager.KW_LISTENER_RESOURCE,
        FlaskManager.KW_EMITTER_RESOURCE,
        FlaskManager.KW_REPOSITORY_RESOURCE,
        FlaskManager.KW_VALIDATOR_RESOURCE,
        FlaskManager.KW_MAPPER_RESOURCE,
        FlaskManager.KW_HELPER_RESOURCE,
        FlaskManager.KW_CONVERTER_RESOURCE
    ]

    # act
    # print(resourceNameCorrectOrder)
    # print(FlaskManager.KW_RESOURCE_LIST)

    # assert
    assert len(resourceNameCorrectOrder) == len(FlaskManager.KW_RESOURCE_LIST)
    for index, resourceName in enumerate(resourceNameCorrectOrder) :
        assert resourceName == FlaskManager.KW_RESOURCE_LIST[index]



import globals
import json

from python_helper import Constant as c
from python_helper import log, Function, ReflectionHelper, ObjectHelper, SettingHelper, EnvironmentHelper, StringHelper

from python_framework.api.src.annotation.EnumAnnotation import EnumItem
from python_framework.api.src.annotation.GlobalExceptionAnnotation import EncapsulateItWithGlobalException
from python_framework.api.src.constant import ConfigurationKeyConstant, JwtConstant
from python_framework.api.src.domain import HttpDomain
from python_framework.api.src.constant import StaticConstant
from python_framework.api.src.constant import LogConstant
from python_framework.api.src.util import FlaskUtil
from python_framework.api.src.util import Serializer
from python_framework.api.src.converter.static import ConverterStatic
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service import WebBrowser
from python_framework.api.src.service.ExceptionHandler import GlobalException
from python_framework.api.src.service import ExceptionHandler
from python_framework.api.src.service import SqlAlchemyProxy
from python_framework.api.src.service import SchedulerManager
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service import ApiKeyManager
from python_framework.api.src.service import SecurityManager
from python_framework.api.src.service.openapi import OpenApiManager


KW_URL = 'url'
KW_DEFAULT_URL = 'defaultUrl'
KW_MODEL = 'model'
KW_API = 'api'
KW_APP = 'app'

KW_METHOD = 'method'

KW_RESOURCE = 'resource'

PYTHON_FRAMEWORK_MODULE_NAME = 'python_framework'
PYTHON_FRAMEWORK_INTERNAL_MODULE_NAME_LIST = [
    PYTHON_FRAMEWORK_MODULE_NAME,
    'TestApi',
    'DevTestApi',
    'LocalTestApi',
    'SecurityManagerTestApi',
    'ApiKeyManagerTestApi',
    'SessionManagerTestApi',
    'SecurityManagerAndApiKeyManagerAndSessionManagerTestApi',
    'ClientTestApi'
]
KW_CONTROLLER_RESOURCE = 'Controller'
KW_SCHEDULER_RESOURCE = 'Scheduler'
KW_SERVICE_RESOURCE = 'Service'
KW_CLIENT_RESOURCE = 'Client'
KW_LISTENER_RESOURCE = 'Listener'
KW_EMITTER_RESOURCE = 'Emitter'
KW_REPOSITORY_RESOURCE = 'Repository'
KW_VALIDATOR_RESOURCE = 'Validator'
KW_MAPPER_RESOURCE = 'Mapper'
KW_HELPER_RESOURCE = 'Helper'
KW_CONVERTER_RESOURCE = 'Converter'
PYTHON_FRAMEWORK_RESOURCE_NAME_DICTIONARY = {
    KW_CONTROLLER_RESOURCE : [
        'ActuatorHealthController',
        'DocumentationController'
    ],
    KW_SCHEDULER_RESOURCE : [],
    KW_SERVICE_RESOURCE : [
        'ActuatorHealthService',
        'DocumentationService'
    ],
    KW_CLIENT_RESOURCE : [],
    KW_LISTENER_RESOURCE : [],
    KW_EMITTER_RESOURCE : [],
    KW_REPOSITORY_RESOURCE : [
        'ActuatorHealthRepository',
        'DocumentationRepository'
    ],
    KW_VALIDATOR_RESOURCE : [],
    KW_MAPPER_RESOURCE : [],
    KW_HELPER_RESOURCE : [],
    KW_CONVERTER_RESOURCE : [
        'ActuatorHealthConverter'
    ]
}
KW_RESOURCE_LIST = list(PYTHON_FRAMEWORK_RESOURCE_NAME_DICTIONARY.keys())


def raiseBadResponseImplementation(cause):
    raise Exception(f'Bad response implementation. {cause}')


def newApp(
    filePath
    , successStatus = True
    , errorStatus = True
    , failureStatus = True
    , warningStatus = True
    , settingStatus = True
    , statusStatus = True
    , infoStatus = True
    , debugStatus = False
    , wrapperStatus = False
    , testStatus = False
    , logStatus = False
):
    globalsInstance = globals.newGlobalsInstance(
        filePath
        , successStatus = successStatus
        , errorStatus = errorStatus
        , failureStatus = failureStatus
        , settingStatus = settingStatus
        , statusStatus = statusStatus
        , infoStatus = infoStatus
        , debugStatus = debugStatus
        , warningStatus = warningStatus
        , wrapperStatus = wrapperStatus
        , testStatus = testStatus
        , logStatus = logStatus
    )
    try:
        app = globals.importResource(
            KW_APP,
            resourceModuleName = ConverterStatic.getValueOrDefault(
                globalsInstance.apiName,
                StringHelper.join(EnvironmentHelper.listDirectoryContent(f'{globalsInstance.BASE_API_PATH}')[0].split(c.DOT)[:-1],character = c.DOT)
            ),
            required = True
        )
    except Exception as exception:
        apiPath = f'{c.DOT}{EnvironmentHelper.OS_SEPARATOR}{globalsInstance.BASE_API_PATH}{globalsInstance.apiName}.py'
        errorMessage = f"Not possible to load app. Make shure it's name is properlly configured at '{globalsInstance.settingFilePath}' and it's instance is named 'app' at '{apiPath}'"
        log.error(newApp, errorMessage, exception)
        raise exception
    if ObjectHelper.isNone(app):
        app = globals.importResource(KW_APP, resourceModuleName=globalsInstance.apiName)
        raise Exception('Not possible to load app. Check logs for more details')
    return app


@Function
def initialize(
    apiInstance
    , defaultUrl = None
    , openInBrowser = False
):
    innerDefaultUrl = getInternalUrl(apiInstance)
    if defaultUrl :
        innerDefaultUrl = f'{innerDefaultUrl}{defaultUrl}'
    def inBetweenFunction(function,*argument,**keywordArgument):
        log.wrapper(initialize, f'''{function.__name__} method''')
        if (openInBrowser):
            log.debug(initialize, f'''Openning "{innerDefaultUrl}" url in rowser''')
            # WebBrowser.openUrlInChrome(innerDefaultUrl)
            WebBrowser.openUrl(innerDefaultUrl)
        def innerFunction(*args,**kwargs):
            try :
                functionReturn = function(*args,**kwargs)
            except Exception as exception :
                raise Exception(f'Failed to initialize. Cause: {str(exception)}')
            return functionReturn
        return innerFunction
    return inBetweenFunction


def runApi(*args, api=None, debug=False, **kwargs):
    if ObjectHelper.isNone(api):
        api = FlaskUtil.getApi()
    muteLogs(api, debug)
    if 'host' not in kwargs and api.host :
        kwargs['host'] = api.host if not 'localhost' == api.host else '0.0.0.0'
    if 'port' not in kwargs and api.port :
        kwargs['port'] = api.port
    log.success(runApi, f'Api will run at {api.internalUrl}')
    log.success(runApi, f'Health check will be available at {api.healthCheckUrl}')
    log.success(runApi, f'Documentation will be available at {api.documentationUrl}')
    log.success(runApi, f'Api static content will be available at {api.staticUrl}')
    for manager in api.managerList:
        manager.onRun(api, api.app)
    api.app.run(*args, debug=debug, **kwargs)
    for manager in api.managerList:
        manager.onShutdown(api, api.app)
    SessionManager.onShutdown(api, api.app)
    ApiKeyManager.onShutdown(api, api.app)
    SecurityManager.onShutdown(api, api.app)
    SchedulerManager.onShutdown(api, api.app)
    SqlAlchemyProxy.onShutdown(api, api.app)
    log.success(runApi, f'{api.globals.apiName} is successfully shuting down')


@Function
def getInternalUrl(apiInstance):
    apiUrl = None
    try :
        apiUrl = f'{apiInstance.scheme}://{apiInstance.host}{c.BLANK if ObjectHelper.isEmpty(apiInstance.port) else f"{c.COLON}{apiInstance.port}"}{apiInstance.baseUrl}'
    except Exception as exception :
        log.error(getInternalUrl, 'Not possible to parse api url', exception)
    return apiUrl


@Function
def muteLogs(apiInstance, debug):
    if not debug:
        import logging
        from werkzeug.serving import WSGIRequestHandler
        werkzeug_logger = logging.getLogger('werkzeug')
        werkzeug_logger.disabled = True
        apiInstance.app.logger.disabled = True
        apschedulerLoggerEnabled = apiInstance.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_ENABLE)
        apscheduler_logger = logging.getLogger('apscheduler.scheduler')
        default_apscheduler_logger = logging.getLogger('apscheduler.executors.default')
        apscheduler_logger.disabled = True if ObjectHelper.isNone(apschedulerLoggerEnabled) else not apschedulerLoggerEnabled
        apscheduler_logger.propagate = not apscheduler_logger.disabled
        default_apscheduler_logger.disabled = apscheduler_logger.disabled
        default_apscheduler_logger.propagate = not apscheduler_logger.disabled
        WSGIRequestHandler.log = lambda self, type, message, *args: None ###- getattr(werkzeug_logger, type)('%s %s' % (self.address_string(), message % args))


@Function
def getRequestBodyAsJson(contentType, requestClass):
    try :
        if OpenApiManager.DEFAULT_CONTENT_TYPE == contentType :
            requestBodyAsJson = FlaskUtil.safellyGetJson()
        elif OpenApiManager.MULTIPART_X_MIXED_REPLACE in contentType :
            requestBodyAsJson = FlaskUtil.safellyGetData()
        else :
            raise Exception(f'Content type "{contentType}" not implemented')
    except Exception as exception :
        raise GlobalException(message='Not possible to parse the request', logMessage=str(exception), status=HttpStatus.BAD_REQUEST)
    validateBodyAsJson(requestBodyAsJson, requestClass)
    return requestBodyAsJson

@Function
def validateBodyAsJson(requestBodyAsJson, requestClass):
    if ObjectHelper.isNotNone(requestClass):
        requestBodyAsJsonIsList = ObjectHelper.isList(requestBodyAsJson)
        requestClassIsList = ObjectHelper.isList(requestClass) and ObjectHelper.isList(requestClass[0])
        if not ((requestBodyAsJsonIsList and requestClassIsList) or (not requestBodyAsJsonIsList and not requestClassIsList)):
            raise GlobalException(message='Bad request', logMessage='Bad request', status=HttpStatus.BAD_REQUEST)

def handleAnyControllerMethodRequest(
    args,
    kwargs,
    contentType,
    resourceInstance,
    resourceInstanceMethod,
    contextRequired,
    apiKeyRequired,
    roleRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    logRequest,
    muteStacktraceOnBusinessRuleException
):
    if ObjectHelper.isNotEmptyCollection(roleRequired):
        return handleSecuredControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            apiKeyRequired,
            roleRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    elif ObjectHelper.isNotEmptyCollection(apiKeyRequired):
        return handleLockedByApiKeyControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            apiKeyRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    elif ObjectHelper.isNotEmptyCollection(contextRequired):
        return handleSessionedControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    return handleControllerMethod(
        args,
        kwargs,
        contentType,
        resourceInstance,
        resourceInstanceMethod,
        requestHeaderClass,
        requestParamClass,
        requestClass,
        logRequest,
        muteStacktraceOnBusinessRuleException
    )

@EncapsulateItWithGlobalException(message=JwtConstant.UNAUTHORIZED_MESSAGE, status=HttpStatus.UNAUTHORIZED)
@SecurityManager.jwtAccessRequired
def handleSecuredControllerMethod(
    args,
    kwargs,
    contentType,
    resourceInstance,
    resourceInstanceMethod,
    contextRequired,
    apiKeyRequired,
    roleRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    logRequest,
    muteStacktraceOnBusinessRuleException
):
    contextList = SecurityManager.getContext()
    if not any(role in set(contextList) for role in roleRequired):
        raise GlobalException(
            message = 'Role not allowed',
            logMessage = f'''Roles {contextList} trying to access denied resource. Allowed roles {roleRequired}''',
            status = HttpStatus.FORBIDDEN
        )
    elif ObjectHelper.isNotEmptyCollection(apiKeyRequired):
        return handleLockedByApiKeyControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            apiKeyRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    elif ObjectHelper.isNotEmptyCollection(contextRequired):
        return handleSessionedControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    return handleControllerMethod(
        args,
        kwargs,
        contentType,
        resourceInstance,
        resourceInstanceMethod,
        requestHeaderClass,
        requestParamClass,
        requestClass,
        logRequest,
        muteStacktraceOnBusinessRuleException
    )

@EncapsulateItWithGlobalException(message=JwtConstant.INVALID_API_KEY_MESSAGE, status=HttpStatus.UNAUTHORIZED)
@ApiKeyManager.jwtAccessRequired
def handleLockedByApiKeyControllerMethod(
    args,
    kwargs,
    contentType,
    resourceInstance,
    resourceInstanceMethod,
    contextRequired,
    apiKeyRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    logRequest,
    muteStacktraceOnBusinessRuleException
):
    contextList = ApiKeyManager.getContext()
    if not any(apiKey in set(contextList) for apiKey in apiKeyRequired):
        raise GlobalException(
            message = 'ApiKey not allowed',
            logMessage = f'''ApiKey {contextList} trying to access denied resource. Allowed apiKeys {apiKeyRequired}''',
            status = HttpStatus.FORBIDDEN
        )
    elif ObjectHelper.isNotEmptyCollection(contextRequired):
        return handleSessionedControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            contextRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )
    return handleControllerMethod(
        args,
        kwargs,
        contentType,
        resourceInstance,
        resourceInstanceMethod,
        requestHeaderClass,
        requestParamClass,
        requestClass,
        logRequest,
        muteStacktraceOnBusinessRuleException
    )

@EncapsulateItWithGlobalException(message=JwtConstant.INVALID_SESSION_MESSAGE, status=HttpStatus.UNAUTHORIZED)
@SessionManager.jwtAccessRequired
def handleSessionedControllerMethod(
    args,
    kwargs,
    contentType,
    resourceInstance,
    resourceInstanceMethod,
    contextRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    logRequest,
    muteStacktraceOnBusinessRuleException
):
    contextList = SessionManager.getContext()
    if not any(context in set(contextList) for context in contextRequired):
        raise GlobalException(
            message = 'Session not allowed',
            logMessage = f'''Sessions {contextList} trying to access denied resource. Allowed contexts: {contextRequired}''',
            status = HttpStatus.FORBIDDEN
        )
    else:
        return handleControllerMethod(
            args,
            kwargs,
            contentType,
            resourceInstance,
            resourceInstanceMethod,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            logRequest,
            muteStacktraceOnBusinessRuleException
        )

@EncapsulateItWithGlobalException()
def handleControllerMethod(
    args,
    kwargs,
    contentType,
    resourceInstance,
    resourceInstanceMethod,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    logRequest,
    muteStacktraceOnBusinessRuleException
):
    requestBodyAsJson = {}
    if resourceInstanceMethod.__name__ in OpenApiManager.ABLE_TO_RECIEVE_BODY_LIST and requestClass :
        requestBodyAsJson = getRequestBodyAsJson(contentType, requestClass)
        if Serializer.requestBodyIsPresent(requestBodyAsJson):
            requestBodyAsJsonSerialized = Serializer.convertFromJsonToObject(requestBodyAsJson, requestClass)
            args = getArgsWithSerializerReturnAppended(args, requestBodyAsJsonSerialized)
    headers = FlaskUtil.addToKwargs(FlaskUtil.KW_HEADERS, requestHeaderClass, FlaskUtil.safellyGetHeaders(), kwargs)
    query = FlaskUtil.addToKwargs(FlaskUtil.KW_PARAMETERS, requestParamClass, FlaskUtil.safellyGetArgs(), kwargs)
    try:
        if resourceInstance.logRequest or  logRequest :
            log.prettyJson(
                resourceInstanceMethod,
                LogConstant.CONTROLLER_REQUEST,
                {
                    'headers': headers,
                    # 'query': FlaskUtil.addToKwargs(FlaskUtil.KW_PARAMETERS, requestParamClass, FlaskUtil.safellyGetArgs(), kwargs), ###- safellyGetUrl() returns query param
                    'body': requestBodyAsJson
                },
                condition = True,
                logLevel = log.INFO
            )
    except Exception as exception:
        log.failure(handleControllerMethod, 'Not possible to log request properly', exception)
    return validateAndReturnResponse(handleAdditionalResponseHeadersIfNeeded(resourceInstanceMethod(resourceInstance,*args[1:],**kwargs)))

@Function
def getArgsWithSerializerReturnAppended(args, argument):
    args = [arg for arg in args]
    args.append(argument)
    # return [arg for arg in args]
    return args

@Function
def getArgumentInFrontOfArgs(args, argument):
    return [argument, *args]

@Function
def getArgsWithResponseClassInstanceAppended(args, responseClass):
    if responseClass :
        resourceInstance = args[0]
        objectRequest = args[1]
        objectRequestSerialized = Serializer.convertFromObjectToObject(objectRequest, responseClass)
        return getArgsWithSerializerReturnAppended(args, objectRequestSerialized)
    return args

@Function
def getResourceFinalName(resourceInstance, resourceName=None):
    if not resourceName :
        resourceName = resourceInstance.__class__.__name__
    for resourceType in KW_RESOURCE_LIST :
        if resourceName.endswith(resourceType):
            resourceName = resourceName[:-len(resourceType)]
            break
    return f'{resourceName[0].lower()}{resourceName[1:]}'

@Function
def getResourceType(resourceInstance, resourceName = None):
    if not resourceName :
        resourceName = resourceInstance.__class__.__name__
    for resourceType in KW_RESOURCE_LIST :
        if resourceName.endswith(resourceType):
            return resourceType

@Function
def setResource(apiInstance, resourceInstance, resourceName=None):
    resourceName = getResourceFinalName(resourceInstance, resourceName=resourceName)
    ReflectionHelper.setAttributeOrMethod(apiInstance,resourceName,resourceInstance)

@Function
def bindResource(apiInstance,resourceInstance):
    FlaskUtil.validateFlaskApi(apiInstance)
    FlaskUtil.validateResourceInstance(resourceInstance)
    setResource(ReflectionHelper.getAttributeOrMethod(apiInstance.resource, getResourceType(resourceInstance).lower()), resourceInstance)

@Function
def validateArgs(args, requestClass, resourceInstanceMethod):
    if ObjectHelper.isNotNone(requestClass):
        resourceInstance = args[0]
        if Serializer.isSerializerList(requestClass):
            if 0 < len(requestClass):
                for index in range(len(requestClass)):
                    if Serializer.isSerializerList(args[index + 1]) and len(args[index + 1]) > 0 :
                        expecteObjectClass = requestClass[index][0]
                        for objectInstance in args[index + 1] :
                            ExceptionHandler.validateArgs(resourceInstance, resourceInstanceMethod, objectInstance, expecteObjectClass)
                    else :
                        objectRequest = args[index + 1]
                        expecteObjectClass = requestClass[index]
                        ExceptionHandler.validateArgs(resourceInstance, resourceInstanceMethod, objectRequest, expecteObjectClass)
        else :
            objectRequest = args[1]
            expecteObjectClass = requestClass
            ExceptionHandler.validateArgs(resourceInstance, resourceInstanceMethod, objectRequest, expecteObjectClass)

def validateKwargs(kwargs, resourceInstance, resourceInstanceMethod, requestHeaderClass, requestParamClass):
    classListToValidate = []
    instanceListToValidate = []
    if ObjectHelper.isNotEmpty(requestHeaderClass):
        classListToValidate.append(requestHeaderClass if ObjectHelper.isNotList(requestHeaderClass) else requestHeaderClass[0])
        instanceListToValidate.append(kwargs.get(FlaskUtil.KW_HEADERS, {}))
    if ObjectHelper.isNotEmpty(requestParamClass):
        classListToValidate.append(requestParamClass if ObjectHelper.isNotList(requestParamClass) else requestParamClass[0])
        instanceListToValidate.append(kwargs.get(FlaskUtil.KW_PARAMETERS, {}))
    validateArgs([resourceInstance, *instanceListToValidate], classListToValidate, resourceInstanceMethod)

@Function
def Controller(
    url = c.SLASH,
    responseHeaders = None,
    tag = 'Tag not defined',
    description = 'Controller not descripted',
    logRequest = False,
    logResponse = False
):
    controllerUrl = url
    controllerTag = tag
    controllerDescription = description
    controllerResponseHeaders = responseHeaders
    controllerLogRequest = logRequest
    controllerLogResponse = logResponse
    def Wrapper(OuterClass,*args,**kwargs):
        log.wrapper(Controller, f'''wrapping {OuterClass.__name__}''', None)
        class InnerClass(OuterClass, FlaskUtil.Resource):
            url = controllerUrl
            responseHeaders = controllerResponseHeaders
            tag = controllerTag
            description = controllerDescription
            logRequest = controllerLogRequest
            logResponse = controllerLogResponse
            def __init__(self,*args,**kwargs):
                log.wrapper(OuterClass, f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})', None)
                apiInstance = FlaskUtil.getApi()
                OuterClass.__init__(self)
                FlaskUtil.Resource.__init__(self,*args,**kwargs)
                self.service = apiInstance.resource.service
                self.globals = apiInstance.globals
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper

@Function
def ControllerMethod(
    url = c.SLASH,
    requestHeaderClass = None,
    requestParamClass = None,
    requestClass = None,
    responseClass = None,
    responseHeaders = None,
    roleRequired = None,
    apiKeyRequired = None,
    contextRequired = None,
    consumes = OpenApiManager.DEFAULT_CONTENT_TYPE,
    produces = OpenApiManager.DEFAULT_CONTENT_TYPE,
    logRequest = False,
    logResponse = False,
    muteStacktraceOnBusinessRuleException = True
):
    controllerMethodUrl = url
    controllerMethodRequestHeaderClass = requestHeaderClass
    controllerMethodRequestParamClass = requestParamClass
    controllerMethodRequestClass = requestClass
    controllerMethodResponseClass = responseClass
    controllerMethodResponseHeaders = responseHeaders
    controllerMethodRoleRequired = roleRequired
    controllerMethodApiKeyRequired = apiKeyRequired
    controllerMethodSessionRequired = contextRequired
    controllerMethodProduces = produces
    controllerMethodConsumes = consumes
    controllerMethodLogRequest = logRequest
    controllerMethodLogResponse = logResponse
    controllerMethodMuteStacktraceOnBusinessRuleException = muteStacktraceOnBusinessRuleException
    def innerMethodWrapper(resourceInstanceMethod,*args,**kwargs):
        log.wrapper(ControllerMethod, f'''wrapping {resourceInstanceMethod.__name__}''', None)
        def innerResourceInstanceMethod(*args,**kwargs):
            f'''(*args, {FlaskUtil.KW_HEADERS}={{}}, {FlaskUtil.KW_PARAMETERS}={{}}, **kwargs)'''
            # r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
            # r.headers["Pragma"] = "no-cache"
            # r.headers["Expires"] = "0"
            # r.headers['Cache-Control'] = 'public, max-age=0'
            resourceInstance = args[0]
            completeResponse = None
            log.info(resourceInstanceMethod, f'{LogConstant.CONTROLLER_SPACE}{FlaskUtil.safellyGetVerb()}{c.SPACE_DASH_SPACE}{FlaskUtil.safellyGetUrl()}')
            try :
                completeResponse = handleAnyControllerMethodRequest(
                    args,
                    kwargs,
                    consumes,
                    resourceInstance,
                    resourceInstanceMethod,
                    contextRequired,
                    apiKeyRequired,
                    roleRequired,
                    requestHeaderClass,
                    requestParamClass,
                    requestClass,
                    logRequest,
                    muteStacktraceOnBusinessRuleException
                )
                validateCompleteResponse(responseClass, completeResponse)
            except Exception as exception :
                log.log(innerResourceInstanceMethod, 'Failure at controller method execution. Getting complete response as exception', exception=exception, muteStackTrace=True)
                completeResponse = getCompleteResponseByException(
                    exception,
                    resourceInstance,
                    resourceInstanceMethod,
                    muteStacktraceOnBusinessRuleException
                )
                ###- request.method:              GET
                ###- request.url:                 http://127.0.0.1:5000/alert/dingding/test?x=y
                ###- request.base_url:            http://127.0.0.1:5000/alert/dingding/test
                ###- request.url_charset:         utf-8
                ###- request.url_root:            http://127.0.0.1:5000/
                ###- str(request.url_rule):       /alert/dingding/test
                ###- request.host_url:            http://127.0.0.1:5000/
                ###- request.host:                127.0.0.1:5000
                ###- request.script_root:
                ###- request.path:                /alert/dingding/test
                ###- request.full_path:           /alert/dingding/test?x=y
                ###- request.args:                ImmutableMultiDict([('x', 'y')])
                ###- request.args.get('x'):       y
            try:
                status = HttpStatus.map(completeResponse[-1])
                additionalResponseHeaders = completeResponse[1]
                if ObjectHelper.isNotNone(resourceInstance.responseHeaders):
                    additionalResponseHeaders = {**resourceInstance.responseHeaders, **additionalResponseHeaders}
                if ObjectHelper.isNotNone(responseHeaders):
                    additionalResponseHeaders = {**responseHeaders, **additionalResponseHeaders}
                responseBody = completeResponse[0] if ObjectHelper.isNotNone(completeResponse[0]) else {'message' : status.enumName}
                httpResponse = FlaskUtil.buildHttpResponse(additionalResponseHeaders, responseBody, status.enumValue, produces)
            except Exception as exception:
                log.failure(innerResourceInstanceMethod, f'Failure while parsing complete response: {completeResponse}. Returning simplified version of it', exception, muteStackTrace=True)
                completeResponse = getCompleteResponseByException(
                    Exception('Not possible to handle complete response'),
                    resourceInstance,
                    resourceInstanceMethod,
                    muteStacktraceOnBusinessRuleException
                )
                httpResponse = FlaskUtil.buildHttpResponse(completeResponse[1], completeResponse[0], completeResponse[-1].enumValue, produces)

            try:
                if resourceInstance.logResponse or logResponse :
                    log.prettyJson(
                        resourceInstanceMethod,
                        LogConstant.CONTROLLER_RESPONSE,
                        {
                            'headers': FlaskUtil.safellyGetResponseHeaders(httpResponse),
                            'body': FlaskUtil.safellyGetFlaskResponseJson(httpResponse), ###- json.loads(Serializer.jsonifyIt(responseBody))
                            'status': status
                        },
                        condition = True,
                        logLevel = log.INFO
                    )
            except Exception as exception:
                log.failure(innerResourceInstanceMethod, 'Not possible to log response properly', exception)

            return httpResponse
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceInstanceMethod)
        innerResourceInstanceMethod.url = controllerMethodUrl
        innerResourceInstanceMethod.requestHeaderClass = controllerMethodRequestHeaderClass
        innerResourceInstanceMethod.requestParamClass = controllerMethodRequestParamClass
        innerResourceInstanceMethod.requestClass = controllerMethodRequestClass
        innerResourceInstanceMethod.responseClass = controllerMethodResponseClass
        innerResourceInstanceMethod.responseHeaders = controllerMethodResponseHeaders
        innerResourceInstanceMethod.roleRequired = controllerMethodRoleRequired
        innerResourceInstanceMethod.apiKeyRequired = controllerMethodApiKeyRequired
        innerResourceInstanceMethod.contextRequired = controllerMethodSessionRequired
        innerResourceInstanceMethod.produces = controllerMethodProduces
        innerResourceInstanceMethod.consumes = controllerMethodConsumes
        innerResourceInstanceMethod.logRequest = controllerMethodLogRequest
        innerResourceInstanceMethod.logResponse = controllerMethodLogResponse
        innerResourceInstanceMethod.muteStacktraceOnBusinessRuleException = controllerMethodMuteStacktraceOnBusinessRuleException
        return innerResourceInstanceMethod
    return innerMethodWrapper


def getAndPersistGlobalException(
    exception,
    resourceInstance,
    resourceInstanceMethod,
    apiInstance = None,
    context = HttpDomain.CONTROLLER_CONTEXT
):
    return ExceptionHandler.handleLogErrorException(
        exception,
        resourceInstance,
        resourceInstanceMethod,
        context,
        apiInstance = apiInstance if ObjectHelper.isNotNone(apiInstance) else FlaskUtil.getNullableApi()
    )


def raiseAndPersistGlobalException(exception, resourceInstance, resourceInstanceMethod, context=None):
    raise getAndPersistGlobalException(exception, resourceInstance, resourceInstanceMethod, context=context)


def getCompleteResponseByException(
    exception,
    resourceInstance,
    resourceInstanceMethod,
    muteStacktraceOnBusinessRuleException
):
    try:
        exception = getAndPersistGlobalException(exception, resourceInstance, resourceInstanceMethod)
        completeResponse = (ExceptionHandler.getDefaultBodyException(exception=exception), {}, exception.status)
        try :
            logErrorMessage = f'Error processing {resourceInstance.__class__.__name__}.{resourceInstanceMethod.__name__} request'
            if HttpStatus.INTERNAL_SERVER_ERROR <= HttpStatus.map(exception.status):
                log.error(resourceInstance.__class__, logErrorMessage, exception)
            else :
                log.failure(resourceInstance.__class__, logErrorMessage, exception=exception, muteStackTrace=muteStacktraceOnBusinessRuleException)
        except Exception as logErrorMessageException :
            log.debug(getCompleteResponseByException, 'Error logging exception at controller', exception=logErrorMessageException)
            log.error(getCompleteResponseByException, 'Error processing request', exception)
        return validateAndReturnResponse(handleAdditionalResponseHeadersIfNeeded(completeResponse))
    except Exception as unexpectedException:
        log.error(getCompleteResponseByException, 'Error while building exception return', unexpectedException)
        log.error(getCompleteResponseByException, 'Error processing request', exception)
        return validateAndReturnResponse((ExceptionHandler.getDefaultBodyException(), {}, ExceptionHandler.DEFAULT_STATUS))

def validateAndReturnResponse(completeResponse):
    if (
        ObjectHelper.isNotTuple(completeResponse) or not 3 == len(completeResponse)
    ) or (
        ObjectHelper.isNotDictionary(completeResponse[1])
    ) or (
        not isinstance(HttpStatus.map(completeResponse[-1]), EnumItem)
    ):
        log.debug(validateAndReturnResponse, f'Invalid completeResponse: {completeResponse}')
        raise Exception('Invalid complete response')
    return completeResponse

def handleAdditionalResponseHeadersIfNeeded(completeResponse):
    # log.log(handleAdditionalResponseHeadersIfNeeded, f'Complete response: {completeResponse}')
    if ObjectHelper.isTuple(completeResponse):
        if 3 == len(completeResponse):
            if ObjectHelper.isTuple(completeResponse[0]):
                ###- ((serviceResponse, serviceHeader), controllerHeader, status)
                return completeResponse[0][0], {**completeResponse[1][0], **completeResponse[0][1]}, completeResponse[1][1]
            return completeResponse
        elif 2 == len(completeResponse):
            if ObjectHelper.isTuple(completeResponse[0]) and 2 == len(completeResponse[0]):
                ###- ((serviceResponse, serviceHeader), status)
                return completeResponse[0][0], completeResponse[0][1], completeResponse[1]
            elif ObjectHelper.isTuple(completeResponse[1]) and 2 == len(completeResponse[1]):
                ###- (serviceResponse, (serviceHeader, status)) --> this case should never happens
                return completeResponse[0], completeResponse[1][0], completeResponse[1][1]
            else:
                ###- (serviceResponse, status) --> missing header
                return completeResponse[0], dict(), completeResponse[1]
    elif ObjectHelper.isList(completeResponse) and 2 == len(completeResponse):
        ###- it can only be guessed at this point
        ###- (serviceResponse, status) --> missing header
        return completeResponse[0], dict(), completeResponse[1]
    ###- totally lost at this point
    return completeResponse


def validateCompleteResponse(responseClass, completeResponse):
    if isNotPythonFrameworkHttpsResponseBody(completeResponse):
        raiseBadResponseImplementation(f'It should be a tuple like this: ({"RESPONSE_CLASS" if ObjectHelper.isNone(responseClass) else responseClass if ObjectHelper.isNotList(responseClass) else responseClass[0]}, HEADERS, HTTPS_CODE). But it is: {completeResponse}')
    if ObjectHelper.isNotNone(responseClass):
        if Serializer.isSerializerList(responseClass):
            if 0 == len(responseClass):
                log.log(validateCompleteResponse, f'"responseClass" was not defined')
            elif 1 == len(responseClass):
                if ObjectHelper.isNotList(responseClass[0])  :
                    if not isinstance(completeResponse[0], responseClass[0]):
                        raiseBadResponseImplementation(f'Response does not match expected class. Expected "{responseClass[0].__name__}", but got "{completeResponse[0].__class__.__name__}"')
                elif ObjectHelper.isNotList(responseClass[0][0]):
                    if ObjectHelper.isNotList(completeResponse[0]):
                        raiseBadResponseImplementation(f'Response is not a list. Expected "{responseClass[0].__class__.__name__}", but found "{completeResponse[0].__class__.__name__}"')
                    elif Serializer.isSerializerList(completeResponse[0]) and 0 < len(completeResponse[0]) and not isinstance(completeResponse[0][0], responseClass[0][0]):
                        raiseBadResponseImplementation(f'Response element class does not match expected element class. Expected "{responseClass[0][0].__name__}", response "{completeResponse[0][0].__class__.__name__}"')
        else :
            if not isinstance(completeResponse[0], responseClass):
                raiseBadResponseImplementation(f'Response does not match expected class. Expected "{responseClass.__name__}", but got "{completeResponse[0].__class__.__name__}"')
    else :
        log.log(validateCompleteResponse, f'"responseClass" was not defined')


def isPythonFrameworkHttpsResponseBody(completeResponse):
    return (
        ObjectHelper.isTuple(completeResponse)
    ) and (
        3 == len(completeResponse)
    ) and (
        isinstance(completeResponse[1], dict)
    ) and (
        isinstance(completeResponse[-1], EnumItem) or isinstance(completeResponse[-1], int)
    )


def isNotPythonFrameworkHttpsResponseBody(completeResponse):
    return not isPythonFrameworkHttpsResponseBody(completeResponse)


def getResourceSelf(apiInstance, resourceType, resourceInstanceName):
    return ReflectionHelper.getAttributeOrMethodByNamePath(
        apiInstance,
        StringHelper.join(
            [
                KW_RESOURCE.lower(),
                resourceType.lower(),
                resourceInstanceName
            ],
            character = c.DOT
        )
    )


@Function
def getGlobals():
    return FlaskUtil.getGlobals()


def getApi():
    return FlaskUtil.getApi()


def getNullableApi():
    return FlaskUtil.getNullableApi()


def defaultResourceInterceptor(*args, **kwargs):
    def defaultInterceptor(resourceInstanceMethod, *defaultInnerArgs, **defaultInnerKwargs):
        def defaultInnerInterceptor(*defaultInnerInterceptorArgs, **defaultInnerInterceptorKwargs):
            # the reason we have to take self out here is because resource annotation adds it back later on
            return resourceInstanceMethod(*defaultInnerInterceptorArgs[1:], **defaultInnerInterceptorKwargs)
        ReflectionHelper.overrideSignatures(defaultInnerInterceptor, resourceInstanceMethod)
        return defaultInnerInterceptor
    return defaultInterceptor



from python_framework.api.src.annotation import EnumAnnotation

from python_framework.api.src.util import UtcDateTimeUtil
from python_framework.api.src.util import FlaskUtil
from python_framework.api.src.util import Serializer
from python_framework.api.src.util import ClientUtil

from python_framework.api.src.service import ExceptionHandler
from python_framework.api.src.service.ExceptionHandler import GlobalException
from python_framework.api.src.service import SessionManager
from python_framework.api.src.service import ApiKeyManager
from python_framework.api.src.service import SecurityManager

from python_framework.api.src.service import SchedulerManager
from python_framework.api.src.service import SqlAlchemyProxy
from python_framework.api.src.service import WebBrowser

from python_framework.api.src.service.openapi import OpenApiManager
from python_framework.api.src.service.openapi import OpenApiDocumentationFile

from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.service.flask import ResourceManager

from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.enumeration.ActuatorHealthStatus import ActuatorHealthStatus
from python_framework.api.src.enumeration.SchedulerType import SchedulerType
from python_framework.api.src.enumeration.WeekDay import WeekDay

from python_framework.api.src.domain import HttpDomain

from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.constant import HealthCheckConstant
from python_framework.api.src.constant import SchedulerConstant
from python_framework.api.src.constant import JwtConstant
from python_framework.api.src.constant import HttpClientConstant
from python_framework.api.src.constant import WeekDayConstant
from python_framework.api.src.constant import ControllerConstant
from python_framework.api.src.constant import StaticConstant
from python_framework.api.src.constant import LogConstant

from python_framework.api.src.converter.static import ConverterStatic

from python_framework.api.src.model import FrameworkModel
from python_framework.api.src.model import ErrorLog
from python_framework.api.src.model import ActuatorHealth

from python_framework.api.src.dto import ActuatorHealthDto
from python_framework.api.src.controller import ActuatorHealthController
from python_framework.api.src.converter import ActuatorHealthConverter
from python_framework.api.src.service import ActuatorHealthService
from python_framework.api.src.repository import ActuatorHealthRepository

from python_framework.api.src.controller import DocumentationController
from python_framework.api.src.service import DocumentationService
from python_framework.api.src.repository import DocumentationRepository

from python_framework.api.src.service.flask.FlaskManager import *

from python_framework.api.src.annotation.EnumAnnotation import *
from python_framework.api.src.annotation.SchedulerAnnotation import *
from python_framework.api.src.annotation.ServiceAnnotation import *
from python_framework.api.src.annotation.RepositoryAnnotation import *
from python_framework.api.src.annotation.client.ClientAnnotation import *
from python_framework.api.src.annotation.client.HttpClientAnnotation import *
from python_framework.api.src.annotation.client.ListenerAnnotation import *
from python_framework.api.src.annotation.client.EmitterAnnotation import *
from python_framework.api.src.annotation.ValidatorAnnotation import *
from python_framework.api.src.annotation.MapperAnnotation import *
from python_framework.api.src.annotation.ConverterAnnotation import *
from python_framework.api.src.annotation.HelperAnnotation import *
from python_framework.api.src.annotation.GlobalExceptionAnnotation import *



from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework.api.src.enumeration.SchedulerType import SchedulerType
from python_framework.api.src.service.flask import FlaskManager
from python_framework.api.src.constant import ConfigurationKeyConstant
from python_framework.api.src.converter.static import ConverterStatic


DEFAUTL_MUTE_LOGS = False
DEFAUTL_DISABLE = False


@Function
def Scheduler(*schedulerArgs, disable=DEFAUTL_DISABLE, muteLogs=DEFAUTL_MUTE_LOGS, **schedulerKwargs) :
    def Wrapper(OuterClass, *args, **kwargs):
        log.wrapper(Scheduler,f'''wrapping {OuterClass.__name__}''')
        class InnerClass(OuterClass):
            def __init__(self,*args,**kwargs):
                log.wrapper(OuterClass,f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                apiInstance = FlaskManager.getApi()
                OuterClass.__init__(self,*args,**kwargs)
                self.globals = apiInstance.globals
                self.service = apiInstance.resource.service
                self.enabled = self.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_ENABLE)
                self.disabled = disable
                self.muteLogs = muteLogs or ConverterStatic.getValueOrDefault(self.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and muteLogs)
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper

@Function
def SchedulerMethod(*methodArgs, requestClass=None, disable=DEFAUTL_DISABLE, muteLogs=DEFAUTL_MUTE_LOGS, **methodKwargs) :
    resourceMethodDisable = disable
    resourceMethodMuteLogs = muteLogs
    def innerMethodWrapper(resourceMethod, *innerMethodArgs, **innerMethodKwargs) :
        log.wrapper(SchedulerMethod,f'''wrapping {resourceMethod.__name__}''')
        apiInstance = FlaskManager.getApi()
        methodClassName = ReflectionHelper.getMethodClassName(resourceMethod)
        methodName = ReflectionHelper.getName(resourceMethod)
        methodKwargs['id'] = methodKwargs.get('id', f'{methodClassName}{c.DOT}{methodName}')
        instancesUpTo = methodKwargs.pop('instancesUpTo', 1)
        weekDays = methodKwargs.pop('weekDays', None)
        resourceMethod.disabled = disable
        resourceMethod.id = methodKwargs['id']
        resourceMethod.muteLogs = muteLogs or ConverterStatic.getValueOrDefault(apiInstance.globals.getApiSetting(ConfigurationKeyConstant.API_SCHEDULER_MUTE_LOGS), DEFAUTL_MUTE_LOGS and resourceMethodMuteLogs)
        if ObjectHelper.isNotEmpty(methodArgs) and SchedulerType.CRON == methodArgs[0] and ObjectHelper.isNotNone(weekDays) and StringHelper.isNotBlank(weekDays) :
            methodKwargs['day_of_week'] = weekDays
        if ObjectHelper.isNotNone(instancesUpTo) :
            methodKwargs['max_instances'] = instancesUpTo
        shedulerArgs = [*methodArgs]
        shedulerKwargs = {**methodKwargs}
        @apiInstance.schedulerManager.task(*shedulerArgs, **shedulerKwargs)
        def innerResourceInstanceMethod(*args, **kwargs) :
            resourceInstanceName = methodClassName[:-len(FlaskManager.KW_SCHEDULER_RESOURCE)]
            resourceInstanceName = f'{resourceInstanceName[0].lower()}{resourceInstanceName[1:]}'
            args = FlaskManager.getArgumentInFrontOfArgs(args, ReflectionHelper.getAttributeOrMethod(apiInstance.resource.scheduler, resourceInstanceName))
            resourceInstance = args[0]
            muteLogs = resourceInstance.muteLogs or resourceMethod.muteLogs
            if resourceInstance.enabled and not resourceInstance.disabled and not resourceMethod.disabled:
                if not muteLogs:
                    log.info(resourceMethod, f'{resourceMethod.id} scheduler method started with args={methodArgs} and kwargs={methodKwargs}')
                methodReturn = None
                try :
                    FlaskManager.validateArgs(args,requestClass,innerResourceInstanceMethod)
                    methodReturn = resourceMethod(*args,**kwargs)
                except Exception as exception :
                    if not muteLogs:
                        log.warning(resourceMethod, f'Not possible to run {resourceMethod.id} properly', exception=exception, muteStackTrace=True)
                    FlaskManager.raiseAndPersistGlobalException(exception, resourceInstance, resourceMethod)
                if not muteLogs:
                    log.info(resourceMethod, f'{resourceMethod.id} scheduler method finished')
                return methodReturn
            if not muteLogs:
                log.warning(resourceMethod, f'{resourceMethod.id} scheduler didn{c.SINGLE_QUOTE}t started. {"Schedulers are disabled" if not resourceInstance.enabled else "This scheduler is disabled" if resourceInstance.disabled else "This scheduler method is disabled"}')
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, resourceMethod)
        resourceMethod.id = methodKwargs.get('id')
        innerResourceInstanceMethod.disable = resourceMethodDisable
        innerResourceInstanceMethod.muteLogs = resourceMethodMuteLogs
        return innerResourceInstanceMethod
    return innerMethodWrapper



import time, requests
from pathlib import Path
from python_helper import Constant as c
from python_helper import Test, ObjectHelper, SettingHelper, EnvironmentHelper, log

### - most likelly to be moved to python_framework library
import subprocess, psutil

CURRENT_PATH = str(Path(__file__).parent.absolute())
ESTIMATED_BUILD_TIME_IN_SECONDS = 6

def killProcesses(givenProcess) :
    try :
        process = psutil.Process(givenProcess.pid)
        for child in process.children(recursive=True):
            child.kill()
        process.kill()
    except Exception as exception :
        log.log(killProcesses, 'Error while killing process', exception=exception)

def getProcess(command, path, timeout=20, muteLogs=False) :
    return subprocess.Popen(
        command, #if not muteLogs else f'''{command} {'2>/dev/null' if EnvironmentHelper.isLinux() else '> NUL'}''',
        cwd = path,
        # stdout = subprocess.PIPE,
        # stderr = None if not muteLogs else subprocess.DEVNULL,
        shell = True
    )


LOG_HELPER_SETTINGS = {
    log.LOG : True,
    log.SUCCESS : True,
    log.SETTING : True,
    log.STATUS : True,
    log.INFO : True,
    log.DEBUG : True,
    log.WARNING : True,
    log.WRAPPER : True,
    log.FAILURE : True,
    log.ERROR : True,
    log.TEST : False,
    log.ENABLE_LOGS_WITH_COLORS: True
}

### - Other possible solutions
# https://stackoverflow.com/questions/4789837/how-to-terminate-a-python-subprocess-launched-with-shell-true

# import subprocess
# import psutil
# def killProcesses(pid):
#     process = psutil.Process(pid)
#     for child in process.children(recursive=True):
#         child.kill()
#     process.kill()
# process = subprocess.Popen(["infinite_app", "param"], shell=True)
# try:
#     process.wait(timeout=3)
# except subprocess.TimeoutExpired:
#     killProcesses(process.pid)
# process = subprocess.Popen(["infinite_app", "param"], shell=True)
# try:
#     process.wait(timeout=3)
# except subprocess.TimeoutExpired:
#     killProcesses(process.pid)


# import os
# import signal
# import subprocess
# The os.setsid() is passed in the argument preexec_fn so
# it's run after the fork() and before  exec() to run the shell.
# pro = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True, preexec_fn=os.setsid)
# os.killpg(os.getpgid(pro.pid), signal.SIGTERM)  # Send the signal to all the process groups


# processPath = f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone'
# command = 'flask run --host=localhost --port=5001'
# commandList = [
#     command + ('2>/dev/null' if EnvironmentHelper.isLinux() else ' > NUL')
# ]
# process = subprocess.Popen(commandList, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=processPath)
# process = subprocess.Popen(
#     commandList,
#     cwd = processPath,
#     stdout = subprocess.PIPE,
#     stderr = subprocess.DEVNULL,
#     shell = True,
#     preexec_fn = EnvironmentHelper.OS.setsid
# )
# time.sleep(3)

GET_ACTUATOR_HEALTH_CONTROLLER = f'/health'
TEST_CONTROLLER = '/test-controller'
TEST_BATCH_CONTROLLER = f'{TEST_CONTROLLER}/batch'

GET_ACTUATOR_HEALTH_CONTROLLER_TEST = '/test/actuator/health'

GET_NONE_ONE = f'{TEST_CONTROLLER}/all-fine-if-its-none'
POST_PAYLOAD_ONE = f'{TEST_CONTROLLER}/payload-validation-test'
GET_NONE_ONE_BATCH = f'{TEST_BATCH_CONTROLLER}/all-fine-if-its-none'
POST_PAYLOAD_ONE_BATCH = f'{TEST_BATCH_CONTROLLER}/payload-validation-test'

def debugRequests() :
    import requests
    import logging
    import http.client
    http.client.HTTPConnection.debuglevel = 1
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True

@Test(environmentVariables={
    SettingHelper.ACTIVE_ENVIRONMENT : 'local'
})
def appRun_whenEnvironmentIsLocalFromLocalConfig_withSuccess() :
    # arrange
    muteLogs = False
    serverPort = 5001
    process = getProcess(
        f'flask run --host=localhost --port={serverPort}',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        BASE_URL = f'http://localhost:{serverPort}/local-test-api'
        payload = {'me':'and you'}
        payloadList = [payload]
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache'
        }

        # act
        # assert
        responseGetHealth = requests.get(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER, headers=headers)
        assert ObjectHelper.equals(
            {'status':'UP'},
            responseGetHealth.json()
        )
        assert 200 == responseGetHealth.status_code

        responseDocumentation = requests.get(BASE_URL + '/documentation', headers=headers)
        assert ObjectHelper.equals(
            'Local Test Api',
            responseDocumentation.json().get('info', {}).get('title')
        ), responseDocumentation.json().get('info', {})
        assert ObjectHelper.equals(
            'This is a Local Test Api service',
            responseDocumentation.json().get('info', {}).get('description')
        ), responseDocumentation.json().get('info', {})
        assert ObjectHelper.equals(
            '0.0.1',
            responseDocumentation.json().get('info', {}).get('version')
        ), responseDocumentation.json().get('info', {})
        assert 200 == responseGetHealth.status_code

        responseGetNone = requests.get(BASE_URL + GET_NONE_ONE)
        assert ObjectHelper.equals(
            {'message': 'OK'},
            responseGetNone.json()
        )
        assert 200 == responseGetNone.status_code

        responseGetNoneBatch = requests.post(BASE_URL + GET_NONE_ONE_BATCH, json=payload, headers=headers)
        expectedBodyResponseWithUri = {
            'message': 'Something bad happened. Please, try again later',
            'timestamp': '2021-03-18 21:43:47.299735',
            "uri": "/local-test-api/test-controller/batch/all-fine-if-its-none"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responseGetNoneBatch.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responseGetNoneBatch.json())
        assert 500 == responseGetNoneBatch.status_code

        responsePostSendPayload = requests.post(BASE_URL + POST_PAYLOAD_ONE, json=payload, headers=headers)
        expectedResponsePostSendPayload = {
            'message': 'Bad request',
            'timestamp': '2021-03-18 21:43:47.405736',
            "uri": "/local-test-api/test-controller/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedResponsePostSendPayload,
            responsePostSendPayload.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponsePostSendPayload} should be equals to {responsePostSendPayload.json()}'
        assert 400 == responsePostSendPayload.status_code

        responsePostSendPayloadList = requests.post(BASE_URL + POST_PAYLOAD_ONE, json=payloadList, headers=headers)
        expectedBodyResponseWithUri = {
            'message': 'Something bad happened. Please, try again later',
            'timestamp': '2021-03-18 21:43:47.299735',
            "uri": "/local-test-api/test-controller/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayloadList.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayloadList.json())
        assert 500 == responsePostSendPayloadList.status_code

        responsePostSendPayloadBatch = requests.post(BASE_URL + POST_PAYLOAD_ONE_BATCH, json=payload, headers=headers)
        expectedBodyResponseWithUri = {
            'message': 'Something bad happened. Please, try again later',
            'timestamp': '2021-03-18 21:43:47.299735',
            "uri": "/local-test-api/test-controller/batch/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayloadBatch.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayloadBatch.json())
        assert 500 == responsePostSendPayloadBatch.status_code

        responsePostSendPayloadListBatch = requests.post(BASE_URL + POST_PAYLOAD_ONE_BATCH, json=payloadList, headers=headers)
        expectedResponsePostSendPayloadListBatch = {
            'message': 'Bad request',
            'timestamp': '2021-03-18 21:43:47.767735',
            "uri": "/local-test-api/test-controller/batch/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedResponsePostSendPayloadListBatch,
            responsePostSendPayloadListBatch.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponsePostSendPayloadListBatch} should be equals to {responsePostSendPayloadListBatch.json()}'
        assert 400 == responsePostSendPayloadListBatch.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception

@Test(environmentVariables={
    SettingHelper.ACTIVE_ENVIRONMENT : 'dev'
})
def appRun_whenEnvironmentIsLocalFromDevConfig_withSuccess() :
    # arrange
    muteLogs = False
    serverPort = 5002
    process = getProcess(
        f'flask run --host=localhost --port={serverPort}',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        BASE_URL = f'http://localhost:{serverPort}/dev-test-api'
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)

        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache'
        }

        # act
        responseGetHealth = requests.post(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER_TEST, json={'status': 'UP'})

        # assert
        assert ObjectHelper.equals(
            [{'status':'UP'}],
            responseGetHealth.json()
        ), f"{[{'status':'UP'}]} == {responseGetHealth.json()}"
        assert 200 == responseGetHealth.status_code

        expectedResponseHeaders = {
            'added': 'header',
            'booleanFalse': False,
            'booleanTrue': True,
            'int': 1,
            'otherInt': -34,
            'float': 1.0,
            'otherFloat': 2.3334
        }
        for k, v in expectedResponseHeaders.items():
            assert k in dict(responseGetHealth.headers), k
            assert ObjectHelper.equals(
                str(v),
                dict(responseGetHealth.headers)[k]
            ), (k, str(v), dict(responseGetHealth.headers)[k])

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception

@Test()
def testing_headersAndParams() :
    #arrange
    muteLogs = False
    serverPort = 5010
    process = getProcess(
        f'flask run --host=localhost --port={serverPort}',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        BASE_URL = f'http://localhost:{serverPort}/local-test-api'
        params = {'first': 'first_p', 'second': 'second_p', 'ytu': 'uty', 'ytv': 'vty'}
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache',
            "Pragma" : 'no-cache',
            'Expires' : '0',
            'Cache-Control' : 'public, max-age=0'
        }
        headers = {'firstHeader': 'firstHeader_h', 'secondHeader': 'secondHeader_h', 'ytu': 'uty', 'ytv': 'vty', **headers}
        log.debug(log.debug, f'variant: {EnvironmentHelper.get("URL_VARIANT")}')
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)

        #act
        #assert
        responseComplete = requests.get(BASE_URL + f'/test/actuator/health/{EnvironmentHelper.get("URL_VARIANT")}/abcdef/oi', params=params, headers=headers)
        expectedBodyResponse = {
            "first": "first_p",
            "firstHeader": "firstHeader_h",
            "second": "second_p",
            "secondHeader": "secondHeader_h",
            "status": "abcdefoi"
        }
        assert ObjectHelper.equals(
            expectedBodyResponse,
            responseComplete.json()
        ), (expectedBodyResponse, responseComplete.json())
        assert 200 == responseComplete.status_code

        responseWithoutParams = requests.get(BASE_URL + f'/test/actuator/health/{EnvironmentHelper.get("URL_VARIANT")}/abcdef/oi', headers=headers)
        expectedBodyResponse = {
            "first": None,
            "firstHeader": "firstHeader_h",
            "second": None,
            "secondHeader": "secondHeader_h",
            "status": "abcdefoi"
        }
        assert ObjectHelper.equals(
            expectedBodyResponse,
            responseWithoutParams.json()
        ), (expectedBodyResponse, responseWithoutParams.json())
        assert 200 == responseWithoutParams.status_code

        responseWithoutHeaders = requests.get(BASE_URL + f'/test/actuator/health/{EnvironmentHelper.get("URL_VARIANT")}/abcdef/oi', params=params)
        expectedBodyResponse = {
            "first": "first_p",
            "firstHeader": None,
            "second": "second_p",
            "secondHeader": None,
            "status": "abcdefoi"
        }
        assert ObjectHelper.equals(
            expectedBodyResponse,
            responseWithoutHeaders.json()
        ), (expectedBodyResponse, responseWithoutHeaders.json())
        assert 200 == responseWithoutHeaders.status_code

        responseWithoutHeadersAndWithoutParams = requests.get(BASE_URL + f'/test/actuator/health/{EnvironmentHelper.get("URL_VARIANT")}/abcdef/oi')
        expectedBodyResponse = {
            "first": None,
            "firstHeader": None,
            "second": None,
            "secondHeader": None,
            "status": "abcdefoi"
        }
        assert ObjectHelper.equals(
            expectedBodyResponse,
            responseWithoutHeadersAndWithoutParams.json()
        ), (expectedBodyResponse, responseWithoutHeadersAndWithoutParams.json())
        assert 200 == responseWithoutHeadersAndWithoutParams.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception

@Test(environmentVariables={
    SettingHelper.ACTIVE_ENVIRONMENT : 'client'
    , log.ENABLE_LOGS_WITH_COLORS: True
    # , **LOG_HELPER_SETTINGS
})
def testing_Client() :
    #arrange
    muteLogs = False
    ### - port on python-build config
    process = getProcess(
        f'python app.py',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)
    TEST_VARIANT = EnvironmentHelper.get("URL_VARIANT")
    log.debug(log.debug, f'variant: {TEST_VARIANT}')
    try:
        URL_PARAM = 'abcd'
        OTHER_URL_PARAM = 'efgh'
        BASE_SIMPLE_URL = 'http://localhost:5022/client-test-api'
        BASE_URL = f'{BASE_SIMPLE_URL}/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}'
        BASE_EXCEPTON_URL = f'{BASE_SIMPLE_URL}/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}'
        PARAMS = {
            'someParam': 'given-someParam'
        }
        HEADERS = {
            'someHeader': 'given-someHeader'
        }

        ##################
        ###- getTest
        ##################
        GET_TEST = 'get'
        GET_TEST_URL = f'/{GET_TEST}'
        testResponse = requests.get(f'{BASE_URL}{GET_TEST_URL}', params=PARAMS, headers=HEADERS, timeout=10)
        assert ObjectHelper.isNotNone(testResponse)
        assert ObjectHelper.equals(200, testResponse.status_code)
        assert ObjectHelper.equals(f'headers-{GET_TEST}', dict(testResponse.headers).get(GET_TEST))
        expectedTestResponse = {
            "someBody": URL_PARAM + OTHER_URL_PARAM,
            "someOtherBody": {
                "someHeader": "given-someHeader",
                "someParam": "given-someParam"
            }
        }
        assert ObjectHelper.equals(
            expectedTestResponse,
            testResponse.json()
        ), (expectedTestResponse, testResponse.json())

        testExceptionResponse = requests.get(f'{BASE_EXCEPTON_URL}{GET_TEST_URL}', params=PARAMS, headers=HEADERS, timeout=10)
        assert ObjectHelper.isNotNone(testExceptionResponse)
        assert ObjectHelper.equals(500, testExceptionResponse.status_code)
        assert ObjectHelper.equals(None, dict(testExceptionResponse.headers).get(GET_TEST))
        expectedTestExceptionResponse = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 22:06:33.914222",
            "uri": f"/client-test-api/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}{GET_TEST_URL}"
        }
        assert ObjectHelper.equals(
            expectedTestExceptionResponse,
            testExceptionResponse.json(),
            ignoreKeyList=['timestamp']
        ), (expectedTestExceptionResponse, testExceptionResponse.json())

        ##################
        ###- postTest
        ##################
        POST_TEST = 'post'
        POST_TEST_URL = f'/{POST_TEST}'
        BODY = {
            'someBody': 'someBodyValue',
            'someOtherBody': 2
        }
        testResponse = requests.post(f'{BASE_URL}{POST_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testResponse)
        assert ObjectHelper.equals(201, testResponse.status_code)
        assert ObjectHelper.equals(f'headers-{POST_TEST}', dict(testResponse.headers).get(POST_TEST))
        expectedTestResponse = {
            "someBody": URL_PARAM + OTHER_URL_PARAM,
            "someOtherBody": {
                "someHeader": "given-someHeader",
                "someParam": "given-someParam",
                **BODY
            }
        }
        assert ObjectHelper.equals(
            expectedTestResponse,
            testResponse.json()
        ), (expectedTestResponse, testResponse.json())

        testExceptionResponse = requests.post(f'{BASE_EXCEPTON_URL}{POST_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testExceptionResponse)
        assert ObjectHelper.equals(500, testExceptionResponse.status_code)
        assert ObjectHelper.equals(None, dict(testExceptionResponse.headers).get(POST_TEST))
        expectedTestExceptionResponse = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 22:06:33.914222",
            "uri": f"/client-test-api/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}{POST_TEST_URL}"
        }
        assert ObjectHelper.equals(
            expectedTestExceptionResponse,
            testExceptionResponse.json(),
            ignoreKeyList=['timestamp']
        ), (expectedTestExceptionResponse, testExceptionResponse.json())

        ##################
        ###- putTest
        ##################
        PUT_TEST = 'put'
        PUT_TEST_URL = f'/{PUT_TEST}'
        BODY = {
            'someBody': 'someBodyValue',
            'someOtherBody': 2
        }
        testResponse = requests.put(f'{BASE_URL}{PUT_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testResponse)
        assert ObjectHelper.equals(200, testResponse.status_code)
        assert ObjectHelper.equals(f'headers-{PUT_TEST}', dict(testResponse.headers).get(PUT_TEST))
        expectedTestResponse = {
            "someBody": URL_PARAM + OTHER_URL_PARAM,
            "someOtherBody": {
                "someHeader": "given-someHeader",
                "someParam": "given-someParam",
                **BODY
            }
        }
        assert ObjectHelper.equals(
            expectedTestResponse,
            testResponse.json()
        ), (expectedTestResponse, testResponse.json())

        testExceptionResponse = requests.put(f'{BASE_EXCEPTON_URL}{PUT_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testExceptionResponse)
        assert ObjectHelper.equals(500, testExceptionResponse.status_code)
        assert ObjectHelper.equals(None, dict(testExceptionResponse.headers).get(PUT_TEST))
        expectedTestExceptionResponse = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 22:06:33.914222",
            "uri": f"/client-test-api/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}{PUT_TEST_URL}"
        }
        assert ObjectHelper.equals(
            expectedTestExceptionResponse,
            testExceptionResponse.json(),
            ignoreKeyList=['timestamp']
        ), (expectedTestExceptionResponse, testExceptionResponse.json())

        ##################
        ###- patchTest
        ##################
        PATCH_TEST = 'patch'
        PATCH_TEST_URL = f'/{PATCH_TEST}'
        BODY = {
            'someBody': 'someBodyValue',
            'someOtherBody': 2
        }
        testResponse = requests.patch(f'{BASE_URL}{PATCH_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testResponse)
        assert ObjectHelper.equals(200, testResponse.status_code)
        assert ObjectHelper.equals(f'headers-{PATCH_TEST}', dict(testResponse.headers).get(PATCH_TEST))
        expectedTestResponse = {
            "someBody": URL_PARAM + OTHER_URL_PARAM,
            "someOtherBody": {
                "someHeader": "given-someHeader",
                "someParam": "given-someParam",
                **BODY
            }
        }
        assert ObjectHelper.equals(
            expectedTestResponse,
            testResponse.json()
        ), (expectedTestResponse, testResponse.json())

        testExceptionResponse = requests.patch(f'{BASE_EXCEPTON_URL}{PATCH_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testExceptionResponse)
        assert ObjectHelper.equals(500, testExceptionResponse.status_code)
        assert ObjectHelper.equals(None, dict(testExceptionResponse.headers).get(PATCH_TEST))
        expectedTestExceptionResponse = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 22:06:33.914222",
            "uri": f"/client-test-api/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}{PATCH_TEST_URL}"
        }
        assert ObjectHelper.equals(
            expectedTestExceptionResponse,
            testExceptionResponse.json(),
            ignoreKeyList=['timestamp']
        ), (expectedTestExceptionResponse, testExceptionResponse.json())

        ##################
        ###- deleteTest
        ##################
        DELETE_TEST = 'delete'
        DELETE_TEST_URL = f'/{DELETE_TEST}'
        BODY = {
            'someBody': 'someBodyValue',
            'someOtherBody': 2
        }
        testResponse = requests.delete(f'{BASE_URL}{DELETE_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testResponse)
        assert ObjectHelper.equals(200, testResponse.status_code)
        assert ObjectHelper.equals(f'headers-{DELETE_TEST}', dict(testResponse.headers).get(DELETE_TEST))
        expectedTestResponse = {
            "someBody": URL_PARAM + OTHER_URL_PARAM,
            "someOtherBody": {
                "someHeader": "given-someHeader",
                "someParam": "given-someParam",
                **BODY
            }
        }
        assert ObjectHelper.equals(
            expectedTestResponse,
            testResponse.json()
        ), (expectedTestResponse, testResponse.json())

        testExceptionResponse = requests.delete(f'{BASE_EXCEPTON_URL}{DELETE_TEST_URL}', params=PARAMS, headers=HEADERS, json=BODY, timeout=10)
        assert ObjectHelper.isNotNone(testExceptionResponse)
        assert ObjectHelper.equals(500, testExceptionResponse.status_code)
        assert ObjectHelper.equals(None, dict(testExceptionResponse.headers).get(DELETE_TEST))
        expectedTestExceptionResponse = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 22:06:33.914222",
            "uri": f"/client-test-api/exception/test/{TEST_VARIANT}/{URL_PARAM}/{OTHER_URL_PARAM}{DELETE_TEST_URL}"
        }
        assert ObjectHelper.equals(
            expectedTestExceptionResponse,
            testExceptionResponse.json(),
            ignoreKeyList=['timestamp']
        ), (expectedTestExceptionResponse, testExceptionResponse.json())

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception

@Test(environmentVariables={
    SettingHelper.ACTIVE_ENVIRONMENT : 'local'
})
def pythonRun_worksProperly() :
    #arrange
    muteLogs = False
    devServerPort = 5001 ### - on local config
    process = getProcess(
        f'python app.py',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    BASE_URL = f'http://localhost:{devServerPort}/local-test-api'
    payload = {'me':'and you'}
    payloadList = [payload]
    headers = {
        "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
        'Cache-Control': 'no-cache'
    }
    time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)

    try:
        #act
        #assert
        session = requests.Session()

        responseGetHealth = session.get(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER, headers=headers)
        assert ObjectHelper.equals(
            {'status':'UP'},
            responseGetHealth.json()
        )
        assert 200 == responseGetHealth.status_code

        responseGetNone = session.get(BASE_URL + GET_NONE_ONE)
        assert ObjectHelper.equals(
            {'message': 'OK'},
            responseGetNone.json()
        )
        assert 200 == responseGetNone.status_code

        responseGetNoneBatch = session.post(BASE_URL + GET_NONE_ONE_BATCH, json=payload, headers=headers)
        expectedBodyResponseWithUri = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 23:34:35.554823",
            "uri": "/local-test-api/test-controller/batch/all-fine-if-its-none"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responseGetNoneBatch.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responseGetNoneBatch.json())
        assert 500 == responseGetNoneBatch.status_code

        responsePostSendPayload = session.post(BASE_URL + POST_PAYLOAD_ONE, json=payload, headers=headers)
        expectedBodyResponseWithUri = {
            'message': 'Bad request',
            'timestamp': '2021-11-15 00:19:30.431058',
            'uri': '/local-test-api/test-controller/payload-validation-test'
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayload.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayload.json())
        assert 400 == responsePostSendPayload.status_code

        responsePostSendPayloadList = session.post(BASE_URL + POST_PAYLOAD_ONE, json=payloadList, headers=headers)
        expectedBodyResponseWithUri = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-15 00:16:46.216862",
            "uri": '/local-test-api/test-controller/payload-validation-test'
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayloadList.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayloadList.json())
        assert 500 == responsePostSendPayloadList.status_code

        responsePostSendPayloadBatch = session.post(BASE_URL + POST_PAYLOAD_ONE_BATCH, json=payload, headers=headers)
        expectedBodyResponseWithUri = {
            "message": "Something bad happened. Please, try again later",
            "timestamp": "2021-11-14 23:34:35.554823",
            "uri": "/local-test-api/test-controller/batch/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayloadBatch.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayloadBatch.json())
        assert 500 == responsePostSendPayloadBatch.status_code

        responsePostSendPayloadListBatch = session.post(BASE_URL + POST_PAYLOAD_ONE_BATCH, json=payloadList, headers=headers)
        expectedBodyResponseWithUri = {
            "message": "Bad request",
            "timestamp": "2021-11-15 00:14:04.622693",
            "uri": "/local-test-api/test-controller/batch/payload-validation-test"
        }
        assert ObjectHelper.equals(
            expectedBodyResponseWithUri,
            responsePostSendPayloadListBatch.json(),
            ignoreKeyList=['timestamp']
        ), (expectedBodyResponseWithUri, responsePostSendPayloadListBatch.json())
        assert 400 == responsePostSendPayloadListBatch.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception


@Test(environmentVariables={
    log.ENABLE_LOGS_WITH_COLORS: True,
    SettingHelper.ACTIVE_ENVIRONMENT : 'security-manager'
})
def pythonRun_securityManager() :
    # arrange
    muteLogs = False
    securityServerPort = 5011 ### - on security-manager config
    process = getProcess(
        f'python app.py',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        TEST_VARIANT = EnvironmentHelper.get("URL_VARIANT")
        BASE_URL = f'http://localhost:{securityServerPort}/security-manager-api'
        BASE_URI = f'{BASE_URL}/test/{TEST_VARIANT}/security-manager'
        POST_LOGIN_URI = '/login'
        GET_CONSUME_URI = '/consume'
        GET_CONSUME_AFTER_REFRESH_URI = '/consume/only-after-refresh'
        PATCH_REFRESH_URI = '/refresh'
        PUT_LOGOUT_URI = '/logout'
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)

        # act
        # assert
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache'
        }
        id = time.time()
        payload = {
            'id': id
        }

        responseGetHealth = requests.get(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER, headers=headers)
        assert ObjectHelper.equals(
            {'status':'UP'},
            responseGetHealth.json()
        )
        assert 200 == responseGetHealth.status_code

        responseGetConsumeBeforeLogin = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeLogin = {
            "message": "Unauthorized",
            "timestamp": "2021-11-02 21:47:32.444629",
            "uri": f"/security-manager-api/test/{TEST_VARIANT}/security-manager/consume"
        }
        assert ObjectHelper.isNotNone(responseGetConsumeBeforeLogin)
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeLogin,
            responseGetConsumeBeforeLogin.json(),
            ignoreKeyList=['timestamp']
        )

        responseLogin = requests.post(BASE_URI + POST_LOGIN_URI, json=payload, headers=headers)
        firstAuthorization = responseLogin.json().get('accessToken')
        firstAuthorizationHeaders = responseLogin.headers
        assert ObjectHelper.isNotNone(firstAuthorization)
        headers['Authorization'] = 'Bearer ' + firstAuthorization
        assert ObjectHelper.isNotNone(firstAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(firstAuthorizationHeaders.get('some'))

        responseGetConsumeBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_ROLE"
                ],
                "data": {
                    "some": "data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeRefresh,
            responseGetConsumeBeforeRefresh.json()
        ), f'{expectedResponseGetConsumeBeforeRefresh} should be equals to {responseGetConsumeBeforeRefresh.json()}'

        responseGetConsumeAfterRefreshBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_AFTER_REFRESH_URI, headers=headers)
        assert ObjectHelper.equals({
                "message": "Role not allowed",
                "timestamp": "2021-11-02 21:47:32.444629",
                "uri": f"/security-manager-api/test/{TEST_VARIANT}/security-manager/consume/only-after-refresh"
            },
            responseGetConsumeAfterRefreshBeforeRefresh.json(),
            ignoreKeyList=['timestamp']
        )
        assert 403 == responseGetConsumeAfterRefreshBeforeRefresh.status_code

        responsePatchRefresh = requests.patch(BASE_URI + PATCH_REFRESH_URI, json=payload, headers=headers)
        patchedAuthorizationHeaders = responseLogin.headers
        patchedAuthorization = responsePatchRefresh.json().get('accessToken')
        assert ObjectHelper.isNotNone(patchedAuthorization)
        headers['Authorization'] = 'Bearer ' + patchedAuthorization
        assert ObjectHelper.isNotNone(responsePatchRefresh)
        assert not ObjectHelper.equals(firstAuthorization, patchedAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(patchedAuthorizationHeaders.get('some'))

        responseGetConsumeAfterRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeAfterRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_ROLE",
                    "TEST_ROLE_REFRESH"
                ],
                "data": {
                    "some": "other data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterRefresh,
            responseGetConsumeAfterRefresh.json()
        ), f'{expectedResponseGetConsumeAfterRefresh} should be equals to {responseGetConsumeAfterRefresh.json()}'

        responseLogout = requests.put(BASE_URI + PUT_LOGOUT_URI, json=payload, headers=headers)
        assert ObjectHelper.isNotNone(responseLogout.json())
        expectedResponseLogout = {'message': 'Logged out'}
        assert ObjectHelper.equals(
            expectedResponseLogout,
            responseLogout.json()
        ), f'{expectedResponseLogout} should be equals to {responseLogout.json()}'
        assert 202 == responseLogout.status_code

        responseGetConsumeAfterLogout = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        assert ObjectHelper.isNotNone(responseGetConsumeAfterLogout.json())
        expectedResponseGetConsumeAfterLogout = {
            'message': 'Unauthorized',
            'timestamp': '2021-11-03 01:10:10.876113',
            "uri": f"/security-manager-api/test/{TEST_VARIANT}/security-manager/consume"
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterLogout,
            responseGetConsumeAfterLogout.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeAfterLogout} should be equals to {responseGetConsumeAfterLogout.json()}'
        assert 401 == responseGetConsumeAfterLogout.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception


@Test(environmentVariables={
    log.ENABLE_LOGS_WITH_COLORS: True,
    SettingHelper.ACTIVE_ENVIRONMENT : 'session-manager'
})
def pythonRun_sessionManager() :
    # arrange
    muteLogs = False
    sessionServerPort = 5012 ### - on session-manager config
    process = getProcess(
        f'python app.py',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        TEST_VARIANT = EnvironmentHelper.get("URL_VARIANT")
        BASE_URL = f'http://localhost:{sessionServerPort}/session-manager-api'
        BASE_URI = f'{BASE_URL}/test/{TEST_VARIANT}/session-manager'
        POST_LOGIN_URI = '/login'
        GET_CONSUME_URI = '/consume'
        GET_CONSUME_AFTER_REFRESH_URI = '/consume/only-after-refresh'
        PATCH_REFRESH_URI = '/refresh'
        PUT_LOGOUT_URI = '/logout'
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache'
        }
        id = time.time()
        payload = {
            'id': id
        }

        # act
        # assert
        responseGetHealth = requests.get(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER, headers=headers)
        assert ObjectHelper.equals(
            {'status':'UP'},
            responseGetHealth.json()
        )
        assert 200 == responseGetHealth.status_code

        responseGetConsumeBeforeLogin = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeLogin = {
            "message": "Invalid session",
            "timestamp": "2021-11-02 21:47:32.444629",
            "uri": f"/session-manager-api/test/{TEST_VARIANT}/session-manager/consume"
        }
        assert ObjectHelper.isNotNone(responseGetConsumeBeforeLogin)
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeLogin,
            responseGetConsumeBeforeLogin.json(),
            ignoreKeyList=['timestamp']
        )

        responseLogin = requests.post(BASE_URI + POST_LOGIN_URI, json=payload, headers=headers)
        firstAuthorization = responseLogin.json().get('accessToken')
        firstAuthorizationHeaders = responseLogin.headers
        assert ObjectHelper.isNotNone(firstAuthorization)
        headers['Context'] = 'Bearer ' + firstAuthorization
        assert ObjectHelper.isNotNone(firstAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(firstAuthorizationHeaders.get('some'))

        responseGetConsumeBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_SESSION"
                ],
                "data": {
                    "some": "data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeRefresh,
            responseGetConsumeBeforeRefresh.json()
        ), f'{expectedResponseGetConsumeBeforeRefresh} should be equals to {responseGetConsumeBeforeRefresh.json()}'

        responseGetConsumeAfterRefreshBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_AFTER_REFRESH_URI, headers=headers)
        expectedResponseGetConsumeAfterRefreshBeforeRefresh = {
            "message": "Session not allowed",
            "timestamp": "2021-11-02 21:47:32.444629",
            "uri": f"/session-manager-api/test/{TEST_VARIANT}/session-manager/consume/only-after-refresh"
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterRefreshBeforeRefresh,
            responseGetConsumeAfterRefreshBeforeRefresh.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeAfterRefreshBeforeRefresh} should be equals to {responseGetConsumeAfterRefreshBeforeRefresh.json()}'
        assert 403 == responseGetConsumeAfterRefreshBeforeRefresh.status_code

        responsePatchRefresh = requests.patch(BASE_URI + PATCH_REFRESH_URI, json=payload, headers=headers)
        patchedAuthorizationHeaders = responseLogin.headers
        patchedAuthorization = responsePatchRefresh.json().get('accessToken')
        assert ObjectHelper.isNotNone(patchedAuthorization)
        headers['Context'] = 'Bearer ' + patchedAuthorization
        assert ObjectHelper.isNotNone(responsePatchRefresh)
        assert not ObjectHelper.equals(firstAuthorization, patchedAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(patchedAuthorizationHeaders.get('some'))

        responseGetConsumeAfterRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeAfterRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_SESSION",
                    "TEST_SESSION_REFRESH"
                ],
                "data": {
                    "some": "other data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterRefresh,
            responseGetConsumeAfterRefresh.json()
        ), f'{expectedResponseGetConsumeAfterRefresh} should be equals to {responseGetConsumeAfterRefresh.json()}'

        responseLogout = requests.put(BASE_URI + PUT_LOGOUT_URI, json=payload, headers=headers)
        assert ObjectHelper.isNotNone(responseLogout.json())
        expectedResponseLogout = {'message': 'Session closed'}
        assert ObjectHelper.equals(
            expectedResponseLogout,
            responseLogout.json()
        ), f'{expectedResponseLogout} should be equals to {responseLogout.json()}'
        assert 202 == responseLogout.status_code

        responseGetConsumeAfterLogout = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        assert ObjectHelper.isNotNone(responseGetConsumeAfterLogout.json())
        expectedResponseGetConsumeAfterLogout = {
            'message': 'Invalid session',
            'timestamp': '2021-11-03 01:10:10.876113',
            "uri": f"/session-manager-api/test/{TEST_VARIANT}/session-manager/consume"
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterLogout,
            responseGetConsumeAfterLogout.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeAfterLogout} should be equals to {responseGetConsumeAfterLogout.json()}'
        assert 401 == responseGetConsumeAfterLogout.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception


@Test(environmentVariables={
    log.ENABLE_LOGS_WITH_COLORS: True,
    SettingHelper.ACTIVE_ENVIRONMENT : 'api-key-manager'
})
def pythonRun_apiKeyManager() :
    # arrange
    muteLogs = False
    apiKeyServerPort = 5013 ### - on apiKey-manager config
    process = getProcess(
        f'python app.py',
        f'{CURRENT_PATH}{EnvironmentHelper.OS_SEPARATOR}apitests{EnvironmentHelper.OS_SEPARATOR}testone',
        muteLogs = muteLogs
    )
    try:
        TEST_VARIANT = EnvironmentHelper.get("URL_VARIANT")
        BASE_URL = f'http://localhost:{apiKeyServerPort}/api-key-manager-api'
        BASE_URI = f'{BASE_URL}/test/{TEST_VARIANT}/api-key-manager'
        POST_LOGIN_URI = '/login'
        GET_CONSUME_URI = '/consume'
        GET_CONSUME_AFTER_REFRESH_URI = '/consume/only-after-refresh'
        PATCH_REFRESH_URI = '/refresh'
        PUT_LOGOUT_URI = '/logout'
        time.sleep(ESTIMATED_BUILD_TIME_IN_SECONDS)
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; CrOS x86_64 12871.102.0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/81.0.4044.141 Safari/537.36",
            'Cache-Control': 'no-cache'
        }
        id = time.time()
        payload = {
            'id': id
        }

        # act
        # assert
        responseGetHealth = requests.get(BASE_URL + GET_ACTUATOR_HEALTH_CONTROLLER, headers=headers)
        assert ObjectHelper.equals(
            {'status':'UP'},
            responseGetHealth.json()
        )
        assert 200 == responseGetHealth.status_code

        responseGetConsumeBeforeLogin = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeLogin = {
            "message": "Invalid apiKey",
            "timestamp": "2021-11-02 21:47:32.444629",
            "uri": f"/api-key-manager-api/test/{TEST_VARIANT}/api-key-manager/consume"
        }
        assert ObjectHelper.isNotNone(responseGetConsumeBeforeLogin)
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeLogin,
            responseGetConsumeBeforeLogin.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeBeforeLogin} should be equals to {responseGetConsumeBeforeLogin.json()}'

        responseLogin = requests.post(BASE_URI + POST_LOGIN_URI, json=payload, headers=headers)
        firstAuthorization = responseLogin.json().get('accessToken')
        firstAuthorizationHeaders = responseLogin.headers
        assert ObjectHelper.isNotNone(firstAuthorization)
        headers['Api-Key'] = 'Bearer ' + firstAuthorization
        assert ObjectHelper.isNotNone(firstAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(firstAuthorizationHeaders.get('some'))

        responseGetConsumeBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeBeforeRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_API_KEY"
                ],
                "data": {
                    "some": "data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeBeforeRefresh,
            responseGetConsumeBeforeRefresh.json()
        ), f'{expectedResponseGetConsumeBeforeRefresh} should be equals to {responseGetConsumeBeforeRefresh.json()}'

        responseGetConsumeAfterRefreshBeforeRefresh = requests.get(BASE_URI + GET_CONSUME_AFTER_REFRESH_URI, headers=headers)
        expectedResponseGetConsumeAfterRefreshBeforeRefresh = {
            "message": "ApiKey not allowed",
            "timestamp": "2021-11-02 21:47:32.444629",
            "uri": f"/api-key-manager-api/test/{TEST_VARIANT}/api-key-manager/consume/only-after-refresh"
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterRefreshBeforeRefresh,
            responseGetConsumeAfterRefreshBeforeRefresh.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeAfterRefreshBeforeRefresh} should be equals to {responseGetConsumeAfterRefreshBeforeRefresh.json()}'
        assert 403 == responseGetConsumeAfterRefreshBeforeRefresh.status_code

        responsePatchRefresh = requests.patch(BASE_URI + PATCH_REFRESH_URI, json=payload, headers=headers)
        patchedAuthorizationHeaders = responseLogin.headers
        patchedAuthorization = responsePatchRefresh.json().get('accessToken')
        assert ObjectHelper.isNotNone(patchedAuthorization)
        headers['Api-Key'] = 'Bearer ' + patchedAuthorization
        assert ObjectHelper.isNotNone(responsePatchRefresh)
        assert not ObjectHelper.equals(firstAuthorization, patchedAuthorization)
        assert ObjectHelper.isNotNone(id)
        assert ObjectHelper.isNone(patchedAuthorizationHeaders.get('some'))

        responseGetConsumeAfterRefresh = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        expectedResponseGetConsumeAfterRefresh = {
            "secured": "information",
            "currentUser": {
                "identity": id,
                "context": [
                    "TEST_API_KEY",
                    "TEST_API_KEY_REFRESH"
                ],
                "data": {
                    "some": "other data"
                }
            }
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterRefresh,
            responseGetConsumeAfterRefresh.json()
        ), f'{expectedResponseGetConsumeAfterRefresh} should be equals to {responseGetConsumeAfterRefresh.json()}'

        responseLogout = requests.put(BASE_URI + PUT_LOGOUT_URI, json=payload, headers=headers)
        assert ObjectHelper.isNotNone(responseLogout.json())
        expectedResponseLogout = {'message': 'ApiKey closed'}
        assert ObjectHelper.equals(
            expectedResponseLogout,
            responseLogout.json()
        ), f'{expectedResponseLogout} should be equals to {responseLogout.json()}'
        assert 202 == responseLogout.status_code

        responseGetConsumeAfterLogout = requests.get(BASE_URI + GET_CONSUME_URI, headers=headers)
        assert ObjectHelper.isNotNone(responseGetConsumeAfterLogout.json())
        expectedResponseGetConsumeAfterLogout = {
            'message': 'Invalid apiKey',
            'timestamp': '2021-11-03 01:10:10.876113',
            "uri": f"/api-key-manager-api/test/{TEST_VARIANT}/api-key-manager/consume"
        }
        assert ObjectHelper.equals(
            expectedResponseGetConsumeAfterLogout,
            responseGetConsumeAfterLogout.json(),
            ignoreKeyList=['timestamp']
        ), f'{expectedResponseGetConsumeAfterLogout} should be equals to {responseGetConsumeAfterLogout.json()}'
        assert 401 == responseGetConsumeAfterLogout.status_code

        killProcesses(process)
    except Exception as exception:
        killProcesses(process)
        raise exception



from python_helper import EnvironmentHelper
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.util import Serializer
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod

import ClientTestDto


BASE_CONTROLLER_URL = f'/exception/test/{EnvironmentHelper.get("URL_VARIANT")}'
GET_TEST = 'get'
GET_TEST_URL = f'/{GET_TEST}'
POST_TEST = 'post'
POST_TEST_URL = f'/{POST_TEST}'
PUT_TEST = 'put'
PUT_TEST_URL = f'/{PUT_TEST}'
PATCH_TEST = 'patch'
PATCH_TEST_URL = f'/{PATCH_TEST}'
DELETE_TEST = 'delete'
DELETE_TEST_URL = f'/{DELETE_TEST}'


@Controller(url=f'{BASE_CONTROLLER_URL}/<string:urlParam>/<string:otherUrlParam>', tag='ClientTestController', description='Client Controller test')
class ClientExceptionTestController:

    @ControllerMethod(
        url = GET_TEST_URL,
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        responseClass = ClientTestDto.ClientTestResponseDto,
        logRequest = True,
        logResponse = True
    )
    def get(self, urlParam=None, otherUrlParam=None, params=None, headers=None):
        return self.globals.api.resource.client.exceptionTest.getMethod(
            f'/{urlParam}/{otherUrlParam}{GET_TEST_URL}/all',
            headers = headers,
            params = params
        )[0], {GET_TEST: f'headers-{GET_TEST}'}, HttpStatus.OK

    @ControllerMethod(
        url = POST_TEST_URL,
        requestHeaderClass = ClientTestDto.ClientTestRequestHeaderDto,
        requestParamClass = ClientTestDto.ClientTestRequestParamDto,
        requestClass = ClientTestDto.ClientTestRequestDto,
        responseClass = ClientTestDto.ClientTestResponseDto,
        logRequest = True,
        logResponse = True
    )
    def post(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        return self.globals.api.resource.client.exceptionTest.postMethod(
            f'/{urlParam}/{otherUrlParam}{POST_TEST_URL}/all',
            dto,
            headers = headers,
            params = params
        )[0], {POST_TEST: f'headers-{POST_TEST}'}, HttpStatus.OK

    @ControllerMethod(
        url = PUT_TEST_URL,
        requestHeaderClass = ClientTestDto.ClientTestRequestHeaderDto,
        requestParamClass = ClientTestDto.ClientTestRequestParamDto,
        requestClass = ClientTestDto.ClientTestRequestDto,
        responseClass = ClientTestDto.ClientTestResponseDto,
        logRequest = True,
        logResponse = True
    )
    def put(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        return self.globals.api.resource.client.exceptionTest.putMethod(
            f'/{urlParam}/{otherUrlParam}{PUT_TEST_URL}/all',
            dto,
            headers = headers,
            params = params
        )[0], {PUT_TEST: f'headers-{PUT_TEST}'}, HttpStatus.OK

    @ControllerMethod(
        url = PATCH_TEST_URL,
        requestHeaderClass = ClientTestDto.ClientTestRequestHeaderDto,
        requestParamClass = ClientTestDto.ClientTestRequestParamDto,
        requestClass = ClientTestDto.ClientTestRequestDto,
        responseClass = ClientTestDto.ClientTestResponseDto,
        logRequest = True,
        logResponse = True
    )
    def patch(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        return self.globals.api.resource.client.exceptionTest.patchMethod(
            f'/{urlParam}/{otherUrlParam}{PATCH_TEST_URL}/all',
            dto,
            headers = headers,
            params = params
        )[0], {PATCH_TEST: f'headers-{PATCH_TEST}'}, HttpStatus.OK

    @ControllerMethod(
        url = DELETE_TEST_URL,
        requestHeaderClass = ClientTestDto.ClientTestRequestHeaderDto,
        requestParamClass = ClientTestDto.ClientTestRequestParamDto,
        requestClass = ClientTestDto.ClientTestRequestDto,
        responseClass = ClientTestDto.ClientTestResponseDto,
        logRequest = True,
        logResponse = True
    )
    def delete(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        return self.globals.api.resource.client.exceptionTest.deleteMethod(
            f'/{urlParam}/{otherUrlParam}{DELETE_TEST_URL}/all',
            dto,
            headers = headers,
            params = params
        )[0], {DELETE_TEST: f'headers-{DELETE_TEST}'}, HttpStatus.OK


@Controller(url=f'{BASE_CONTROLLER_URL}', tag='ClientTestController', description='Client Controller test')
class ClientExceptionTestBatchController:

    @ControllerMethod(
        url = f'/<string:urlParam>/<string:otherUrlParam>{GET_TEST_URL}/all',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logRequest = True,
        logResponse = True
    )
    def get(self, urlParam=None, otherUrlParam=None, params=None, headers=None):
        raise Exception("Some unexpected exception")

    @ControllerMethod(
        url = f'/<string:urlParam>/<string:otherUrlParam>{POST_TEST_URL}/all',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logRequest = True,
        logResponse = True
    )
    def post(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        raise Exception("Some unexpected exception")

    @ControllerMethod(
        url = f'/<string:urlParam>/<string:otherUrlParam>{PUT_TEST_URL}/all',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logRequest = True,
        logResponse = True
    )
    def put(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        raise Exception("Some unexpected exception")

    @ControllerMethod(
        url = f'/<string:urlParam>/<string:otherUrlParam>{PATCH_TEST_URL}/all',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logRequest = True,
        logResponse = True
    )
    def patch(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        raise Exception("Some unexpected exception")

    @ControllerMethod(
        url = f'/<string:urlParam>/<string:otherUrlParam>{DELETE_TEST_URL}/all',
        requestHeaderClass = [ClientTestDto.ClientTestRequestHeaderDto],
        requestParamClass = [ClientTestDto.ClientTestRequestParamDto],
        requestClass = [ClientTestDto.ClientTestRequestDto],
        responseClass = [[ClientTestDto.ClientTestResponseDto]],
        logRequest = True,
        logResponse = True
    )
    def delete(self, dto, urlParam=None, otherUrlParam=None, params=None, headers=None):
        raise Exception("Some unexpected exception")



from python_framework.api.src.annotation.ConverterAnnotation import Converter, ConverterMethod
from python_framework.api.src.model import ActuatorHealth
from python_framework.api.src.dto import ActuatorHealthDto

@Converter()
class ActuatorHealthTestConverter:

    @ConverterMethod(
        requestClass=[[ActuatorHealth.ActuatorHealth]],
        responseClass=[[ActuatorHealthDto.ActuatorHealthResponseDto]]
    )
    def fromModelListToResponseDtoList(self, modelList, dtoList) :
        return dtoList



class ClientTestRequestParamDto:
    def __init__(self,
        someParam = None
    ):
        self.someParam = someParam

class ClientTestRequestHeaderDto:
    def __init__(self,
        someHeader = None
    ):
        self.someHeader = someHeader


class ClientTestRequestDto:
    def __init__(self,
        someBody = None,
        someOtherBody = None
    ):
        self.someBody = someBody
        self.someOtherBody = someOtherBody


class ClientTestResponseDto:
    def __init__(self,
        someBody = None,
        someOtherBody = None
    ):
        self.someBody = someBody
        self.someOtherBody = someOtherBody



from python_helper import EnvironmentHelper
from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod
from python_framework.api.src.dto import ActuatorHealthDto

from dto import EnumAsQueryDto

from dto import TestRequestDto

@Controller(url='/test/actuator/health', tag='HealthCheckTest', description='HealthCheck controller test')
class ActuatorHealthTestController:

    @ControllerMethod(
        requestClass = EnumAsQueryDto.EnumAsQueryRequestDto,
        responseClass = [[ActuatorHealthDto.ActuatorHealthResponseDto]],
        logResponse = True
    )
    def post(self, dto):
        return self.service.status.findAllByStatus(dto), HttpStatus.OK

    @ControllerMethod(
        url=f'/{EnvironmentHelper.get("URL_VARIANT")}',
        logRequest = True,
        requestHeaderClass = [TestRequestDto.TestRequestHeaderDto],
        requestParamClass = [TestRequestDto.TestRequestParamDto],
        requestClass = ActuatorHealthDto.ActuatorHealthResponseDto,
        responseClass = [[ActuatorHealthDto.ActuatorHealthResponseDto]],
        logResponse = True
    )
    def put(self, dto, headers=None, params=None):
        # from flask import request
        # print(f'{self.__class__.__name__}.put -> request.headers: {request.headers}')
        # print(f'headers.firstHeader: {headers.firstHeader}')
        # print(f'headers.secondHeader: {headers.secondHeader}')
        # print(f'params.first: {params.first}')
        # print(f'params.second: {params.second}')
        return [dto], HttpStatus.OK

    @ControllerMethod(
        url = f'/{EnvironmentHelper.get("URL_VARIANT")}/<string:status>/<string:secondStatus>',
        requestHeaderClass = [TestRequestDto.TestRequestHeaderDto],
        requestParamClass = [TestRequestDto.TestRequestParamDto],
        responseClass = [TestRequestDto.TestResponseDto],
        logResponse = True
    )
    def get(self, status=None, secondStatus=None, headers=None, params=None):
        return self.globals.api.resource.client.some.getOnActuatorHealth(
            ActuatorHealthDto.ActuatorHealthResponseDto(status+secondStatus),
            headers=headers,
            params=params
        ), HttpStatus.OK



from python_framework.api.src.enumeration.HttpStatus import HttpStatus
from python_framework.api.src.service.flask.FlaskManager import Controller, ControllerMethod

@Controller(url='/test-controller', tag='MyUrl', description='My url controller')
class MyController:

    @ControllerMethod(
        url = '/payload-validation-test',
        requestClass = [[dict]],
        responseClass = [dict],
        logRequest = True,
        logResponse = True,
        muteStacktraceOnBusinessRuleException = False
    )
    def post(self, requestBodyList):
        return requestBodyList, HttpStatus.OK

    @ControllerMethod(
        url = '/payload-validation-test',
        requestClass = [[dict]],
        responseClass = [dict],
        logRequest = True,
        logResponse = True
    )
    def patch(self, requestBodyList):
        return requestBodyList, HttpStatus.OK

    @ControllerMethod(
        url = '/all-fine-if-its-none',
        logRequest = True,
        logResponse = True
    )
    def get(self):
        return None, HttpStatus.OK

@Controller(url='/test-controller/batch', tag='MyUrl', description='My url controller')
class MyBatchController:

    @ControllerMethod(
        url = '/payload-validation-test',
        requestClass = [dict],
        responseClass = [[dict]],
        logRequest = True,
        logResponse = True
    )
    def post(self, requestBodyList):
        return requestBodyList  , HttpStatus.OK

    @ControllerMethod(
        url = '/payload-validation-test',
        requestClass = [dict],
        responseClass = [[dict]],
        logRequest = True,
        logResponse = True
    )
    def patch(self, requestBodyList):
        return requestBodyList, HttpStatus.OK

    @ControllerMethod(
        url = '/all-fine-if-its-none',
        logRequest = True,
        logResponse = True
    )
    def get(self):
        return [None], HttpStatus.OK



from python_framework.api.src.annotation.ServiceAnnotation import Service, ServiceMethod

import EnumAsQueryDto

@Service()
class StatusService :

    @ServiceMethod(requestClass=[EnumAsQueryDto.EnumAsQueryRequestDto])
    def findAllByStatus(self, dto):
        modelList = self.repository.actuatorHealthTest.findAllByStatus(dto.status)
        return self.converter.actuatorHealthTest.fromModelListToResponseDtoList(modelList), {
            'added': 'header',
            'booleanFalse': False,
            'booleanTrue': True,
            'int': 1,
            'otherInt': -34,
            'float': 1.0,
            'otherFloat': 2.3334
        }



class TestRequestParamDto:
    def __init__(self,
        first = None,
        second = None
    ):
        self.first = first
        self.second = second


class TestRequestHeaderDto:
    def __init__(self,
        firstHeader = None,
        secondHeader = None
    ):
        self.firstHeader = firstHeader
        self.secondHeader = secondHeader


class TestResponseDto:
    def __init__(self,
        status = None,
        first = None,
        second = None,
        firstHeader = None,
        secondHeader = None
    ):
        self.status = status
        self.first = first
        self.second = second
        self.firstHeader = firstHeader
        self.secondHeader = secondHeader



from python_helper import Constant as c
from python_helper import log, ReflectionHelper, ObjectHelper, StringHelper

ENUM_VALUE_AS_STRING_KEY = 'enumValue'
ENUM_NAME_AS_STRING_KEY = 'enumName'
ENUM_AS_STRING_KEY = 'enum'

MAP_METHOD_NAME = 'map'
ENUM_STATIC_METHOD_NAME_LIST = [
    MAP_METHOD_NAME
]

class EnumItemStr(str) :
    ...
class EnumItemInt(int) :
    ...
class EnumItemFloat(float) :
    ...
class EnumItemDict(dict) :
    ...
class EnumItemSet(set) :
    ...
class EnumItemTuple(tuple) :
    ...
class EnumItemList(list) :
    ...

class EnumClass(object) :
    ...

def updateEnumItem(enumItem, key, value) :
    enumItem.__enumItemMap__[key] = value
    enumItem.__enumItemEqMap__[key] = str(value)
    ReflectionHelper.setAttributeOrMethod(enumItem, key, ENUM_ITEM_CLASS_DICTIONARY.get(type(value))(value))

class EnumItem :
    def __init__(self, *args, **kwargs):
        self.__enumItemMap__ = {}
        self.__enumItemEqMap__ = {}
        for key, value in kwargs.items():
            updateEnumItem(self, key, value)

    def __get__(self, obj, objtype=None):
        return self if not ReflectionHelper.hasAttributeOrMethod(self, ENUM_VALUE_AS_STRING_KEY) else self.enumValue

    def __str__(self):
        return str(ReflectionHelper.getAttributeOrMethod(self, ENUM_VALUE_AS_STRING_KEY, muteLogs=True))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if isinstance(other, EnumItem):
            return str(other.__enumItemEqMap__) == str(self.__enumItemEqMap__)
        return self.enumValue == other

    def __ne__(self, other):
        if isinstance(other, EnumItem):
            return str(other.__enumItemEqMap__) != str(self.__enumItemEqMap__)
        return self.enumValue != other

    def __lt__(self, other):
        if isinstance(other, EnumItem) and not isinstance(self.enumValue, EnumItem):
            return self.enumValue < other.enumValue
        return self < other

    def __gt__(self, other):
        if isinstance(other, EnumItem) and not isinstance(self.enumValue, EnumItem):
            return self.enumValue > other.enumValue

    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __ge__(self, other):
        return self.__eq__(other) or self.__gt__(other)


ENUM_ITEM_CLASS_DICTIONARY = {
    str : EnumItemStr,
    EnumItemStr : EnumItemStr,
    int : EnumItemInt,
    EnumItemInt : EnumItemInt,
    float : EnumItemFloat,
    EnumItemFloat : EnumItemFloat,
    dict : EnumItemDict,
    EnumItemDict : EnumItemDict,
    set : EnumItemSet,
    EnumItemSet : EnumItemSet,
    tuple : EnumItemTuple,
    EnumItemTuple : EnumItemTuple,
    list : EnumItemList,
    EnumItemList : EnumItemList,
    EnumItem : EnumItem
}

def isEnum(possibleEnum) :
    itIs = False
    try :
        itIs = isinstance(possibleEnum, EnumItem)
    except Exception as exception :
        log.error(isEnum, f'Not possible to evaluate. Returning {itIs} by default', exception)
    return itIs

def isEnumItem(possibleEnumItem) :
    itIs = False
    try :
        itIs = isinstance(possibleEnumItem, EnumItem)
    except Exception as exception :
        log.error(isEnumItem, f'Not possible to evaluate. Returning {itIs} by default', exception)
    return itIs

def Enum(instanceLog=False, associateReturnsTo=ENUM_VALUE_AS_STRING_KEY) :
    def Wrapper(OuterEnum, *args, **kwargs):
        def __raiseBadImplementation__(enumValue, enum=None) :
            raise Exception(f'Not possible to implement "{enumValue}" enum value in {enum} enum')
        def __raiseEnumValueNotImplemented__(enumValue, enumClass=None, enumEqList=None, enumMap=None) :
            raise Exception(f'Not possible to retrieve "{enumValue}" of type: {type(enumValue)} enum value from {ReflectionHelper.getName(enumClass)}: {enumEqList} enum. Enum.__enumMap__ = {enumMap}')
        if instanceLog :
            log.log(Enum,f'''wrapping {OuterEnum.__name__}''')
        class InnerEnum(OuterEnum, EnumClass):
            OuterEnum.__enumMap__ = {}
            OuterEnum.__enumEqList__ = []
            def __init__(self,*args,**kwargs):
                originalClassAttributeValueList = ReflectionHelper.getAttributeNameList(OuterEnum)
                if instanceLog :
                    log.prettyPython(OuterEnum, 'originalClassAttributeValueList', originalClassAttributeValueList)
                OuterEnum.__init__(self,*args,**kwargs)
                self.__associateReturnsTo__ = associateReturnsTo
                attributeDataList = ReflectionHelper.getAttributeDataList(self)
                if instanceLog :
                    log.prettyPython(OuterEnum, 'attributeDataList', attributeDataList)
                for attribute, enumValue in attributeDataList :
                    if enumValue not in ENUM_STATIC_METHOD_NAME_LIST and enumValue not in originalClassAttributeValueList :
                        __raiseBadImplementation__(enumValue, enum=self)
                    nativeClassDataList = []
                    attributeAttributeDataList = ReflectionHelper.getAttributeDataList(attribute)
                    if instanceLog :
                        log.prettyPython(OuterEnum, 'attributeAttributeDataList', attributeAttributeDataList)
                        # print(f'type({attribute}): {type(attribute)}')
                    association = ReflectionHelper.getAttributeOrMethod(attribute, self.__associateReturnsTo__, muteLogs=True)
                    if type(association) is type(None) :
                        # print('type(association) is type(None)')
                        ReflectionHelper.setAttributeOrMethod(attribute, ENUM_VALUE_AS_STRING_KEY, ENUM_ITEM_CLASS_DICTIONARY.get(type(enumValue))(enumValue))
                        nativeClassDataList = ReflectionHelper.getAttributeDataList(type(enumValue)())
                        # print(f'type({enumValue}): {type(enumValue)}, type({type(enumValue)}()): {type(type(enumValue)())}')
                    else :
                        # print('not type(association) is type(None)')
                        ReflectionHelper.setAttributeOrMethod(attribute, ENUM_VALUE_AS_STRING_KEY, ENUM_ITEM_CLASS_DICTIONARY.get(type(association))(association))
                        nativeClassDataList = ReflectionHelper.getAttributeDataList(type(association)())
                        # print(f'type({association}): {type(association)}, type({type(association)}()): {type(type(association)())}')
                    # print()
                    attribute.enumValue.enumName = enumValue
                    ReflectionHelper.setAttributeOrMethod(attribute, ENUM_NAME_AS_STRING_KEY, enumValue)
                    if isinstance(attribute, EnumItem) :
                        # print(attribute, attribute.enumName, attribute)
                        updateEnumItem(attribute, attribute.enumName, attribute) ###- so that __eq__ does not crash
                    if not c.TYPE_FUNCTION == ReflectionHelper.getClassName(attribute) :
                        attribute.enumValue.__enumItemMap__ = attribute.__enumItemMap__
                        attribute.enumValue.__enumItemEqMap__ = attribute.__enumItemEqMap__
                    for dataInstance, dataName in attributeAttributeDataList :
                        nativeData = False
                        for _, name in nativeClassDataList :
                            if dataName == name :
                                nativeData = True
                                break
                        if not nativeData :
                            ReflectionHelper.setAttributeOrMethod(attribute.enumValue, dataName, dataInstance)
                    ReflectionHelper.setAttributeOrMethod(attribute.enumValue, ENUM_VALUE_AS_STRING_KEY, ReflectionHelper.getAttributeOrMethod(attribute, self.__associateReturnsTo__, muteLogs=True))
                    ReflectionHelper.setAttributeOrMethod(attribute.enumValue, ENUM_NAME_AS_STRING_KEY, enumValue)
                    ReflectionHelper.setAttributeOrMethod(attribute.enumValue, ENUM_AS_STRING_KEY, self)
                if ObjectHelper.isEmpty(OuterEnum.__enumEqList__) :
                    for attribute, value in attributeDataList :
                        # print(f'type({value}): {type(value)}')
                        # print(f'type({attribute}): {type(attribute)}')
                        valueAsString = str(attribute.enumValue)
                        if valueAsString not in ENUM_STATIC_METHOD_NAME_LIST :
                            if valueAsString not in self.__enumMap__ :
                                self.__enumMap__[valueAsString] = attribute.enumValue ###- ReflectionHelper.getAttributeOrMethod(attribute, valueAsString, muteLogs=True) ###- attribute
                            else :
                                __raiseBadImplementation__(valueAsString, enum=self)
                            if value not in self.__enumMap__ :
                                self.__enumMap__[value] = attribute
                            elif not self.__enumMap__[value] == attribute :
                                __raiseBadImplementation__(valueAsString, enum=self)
                            self.__enumEqList__.append(f'{value}({self.__associateReturnsTo__}:{valueAsString})')

            def __eq__(self, other) :
                return str(self) == str(other)

            def __ne__(self, other) :
                return str(self) != str(other)

            def __str__(self) :
                return StringHelper.join([ReflectionHelper.getName(OuterEnum), c.COLON_SPACE, str(self.__enumEqList__)])

            def __repr__(self) :
                return self.__str__()

            @staticmethod
            def map(enumItemOrEnumItemValue) :
                if instanceLog :
                    log.log(OuterEnum,f'''enumItemOrEnumItemValue: {enumItemOrEnumItemValue}''')
                if ObjectHelper.isNone(enumItemOrEnumItemValue) :
                    return enumItemOrEnumItemValue
                else :
                    # try :
                    #     print(f'''not ReflectionHelper.hasAttributeOrMethod({enumItemOrEnumItemValue}, 'enumName') --> {str(enumItemOrEnumItemValue)}''')
                    # except :
                    #     print(f'''ReflectionHelper.hasAttributeOrMethod({enumItemOrEnumItemValue}, 'enumName') --> {enumItemOrEnumItemValue.enumName}''')
                    mappedEnum = OuterEnum.__enumMap__.get(str(enumItemOrEnumItemValue) if not ReflectionHelper.hasAttributeOrMethod(enumItemOrEnumItemValue, 'enumName') else enumItemOrEnumItemValue.enumName)
                    return mappedEnum if ObjectHelper.isNotNone(mappedEnum) else __raiseEnumValueNotImplemented__(enumItemOrEnumItemValue, enumClass=OuterEnum, enumEqList=OuterEnum.__enumEqList__, enumMap=OuterEnum.__enumMap__)
        ReflectionHelper.overrideSignatures(InnerEnum, OuterEnum)
        return InnerEnum
    return Wrapper
